from typing import Any, Final

from ..jobs._computation import ComputeCriteria, ComputeEBEs, ComputeFIM
from ..jobs._fitting import Fit
from ..jobs._prediction import PredictSurvLogps, PredictTrajectories, PredictY
from ._defs import Job

# Constants
DEFAULT_HYPERPARAMETERS_FIELDS: Final[tuple[str, ...]] = (
    "max_iterations",
    "n_chains",
    "warmup",
    "n_steps",
)


DEFAULT_HYPERPARAMETERS: Final[dict[type[Job], dict[str, Any]]] = {
    ComputeCriteria: {
        "max_iterations": 100,
        "n_chains": 10,
        "warmup": 200,
        "n_steps": 10,
    },
    ComputeEBEs: {"max_iterations": 100, "n_chains": 10, "warmup": 200, "n_steps": 10},
    ComputeFIM: {"max_iterations": 100, "n_chains": 10, "warmup": 200, "n_steps": 10},
    Fit: {"max_iterations": 500, "n_chains": 10, "warmup": 200, "n_steps": 10},
    PredictSurvLogps: {
        "max_iterations": 100,
        "n_chains": 10,
        "warmup": 200,
        "n_steps": 10,
    },
    PredictTrajectories: {
        "max_iterations": 100,
        "n_chains": 10,
        "warmup": 200,
        "n_steps": 10,
    },
    PredictY: {"max_iterations": 100, "n_chains": 10, "warmup": 200, "n_steps": 10},
}



import itertools

import torch

from ..typedefs._defs import MatRepr, Trajectory


def check_inf(tensors: tuple[tuple[torch.Tensor | None, str], ...]):
    """Check if any of the tensors contain infinity.

    Args:
        tensors (tuple[tuple[torch.Tensor | None, str] ,...]): The tuples to check.

    Raises:
        ValueError: If any of the tensors contain infinity.
    """
    for t, name in tensors:
        if t is not None and t.isinf().any():
            raise ValueError(f"Tensor {name} cannot contain inf values")


def check_nan(tensors: tuple[tuple[torch.Tensor | None, str], ...]):
    """Check if any of the tensors contain NaNs.

    Args:
        tensors (tuple[tuple[torch.Tensor | None, str] ,...]): The tuples to check.

    Raises:
        ValueError: If any of the tensors contain NaNs.
    """
    for t, name in tensors:
        if t is not None and t.isnan().any():
            raise ValueError(f"Tensor {name} cannot contain NaN values")


def check_consistent_size(
    groups: tuple[tuple[torch.Tensor | int | None, int | None, str], ...],
):
    """Checks if all the tensors are consistent in size.

    Args:
        groups (tuple[tuple[torch.Tensor | int | None, int | None, str], ...]):
            The tuples to check.

    Raises:
        ValueError: If any of the sizes are inconsistent.
    """
    for (t0, d0, name0), (t1, d1, name1) in itertools.pairwise(groups):
        if t0 is None or t1 is None:
            continue
        dim0 = t0 if isinstance(t0, int) else t0.size(d0)
        dim1 = t1 if isinstance(t1, int) else t1.size(d1)

        if dim0 != dim1:
            raise ValueError(
                f"{dim0} != {dim1} at dims {d0, d1} for args {name0, name1}"
            )


def check_trajectory_empty(trajectories: list[Trajectory]):
    """Check if the trajectories are not empty.

    Args:
        trajectories (list[Trajectory]): The trajectories.

    Raises:
        ValueError: If some trajectory is empty.
    """
    if any(len(trajectory) == 0 for trajectory in trajectories):
        raise ValueError("Trajectories must not be empty")


def check_trajectory_sorting(trajectories: list[Trajectory]):
    """Check if the trajectories are well sorted.

    Args:
        trajectories (list[Trajectory]): The trajectories.

    Raises:
        ValueError: If some trajectory is not sorted.
    """
    if any(
        not all(t0 <= t1 for t0, t1 in itertools.pairwise(t for t, _ in trajectory))
        for trajectory in trajectories
    ):
        raise ValueError(
            "Trajectories must be sorted by time, in ascending order. Also ensure "
            "there are no NaN values as this will trigger the check"
        )


def check_trajectory_c(trajectories: list[Trajectory], c: torch.Tensor | None):
    """Check if the trajectories are compatible with censoring times.

    Args:
        trajectories (list[Trajectory]): The trajectories.
        c (torch.Tensor | None): The censoring times.

    Raises:
        ValueError: If some trajectory is not compatible with the censoring.
    """
    if c is not None and any(
        trajectory[-1][0] > c for trajectory, c in zip(trajectories, c, strict=True)
    ):
        raise ValueError("Last trajectory time must not be greater than censoring time")


def check_matrix_dim(mat_repr: MatRepr, name: str):
    """Sets dimensions for matrix.

    Args:
        mat_repr (MatRepr): The matrix representation.
        name (str): The matrix name.

    Raises:
        ValueError: If flat tensor is not flat.
        ValueError: If the number of elements is incompatible with method "full".
        ValueError: If the number of elements is incompatible with method "diag".
        ValueError: If the number of elements is not one and the method is "ball".
        ValueError: If the method is not in ("full", "diag", "ball").
    """
    flat, dim, method = mat_repr

    if flat.ndim != 1:
        raise ValueError(
            f"flat must be flat tensor, got shape {flat.shape} for matrix {name}"
        )

    match method:
        case "full":
            if flat.numel() != (dim * (dim + 1)) // 2:
                raise ValueError(
                    f"{flat.numel()} is incompatible with full matrix {name} of "
                    f"dimension {dim}"
                )
        case "diag":
            if flat.numel() != dim:
                raise ValueError(
                    f"{flat.numel()} is incompatible with diag matrix {name} of "
                    f"dimension {dim}"
                )
        case "ball":
            if flat.numel() != 1:
                raise ValueError(
                    f"Expected 1 element for flat, got {flat.numel()} for matrix {name}"
                )
        case _:
            raise ValueError(
                f"Method must be be either 'full', 'diag' or 'ball', got {method}"
            )



from collections.abc import Callable
from typing import Any, cast

import torch

from ..typedefs._defs import AuxData
from ..utils._dtype import get_dtype


class MetropolisHastingsSampler:
    """A robust Metropolis-Hastings sampler with adaptive step size."""

    logpdfs_aux_fn: Callable[[torch.Tensor], tuple[torch.Tensor, AuxData]]
    n_chains: int
    adapt_rate: int | float
    target_accept_rate: int | float
    b: torch.Tensor
    logpdfs: torch.Tensor
    aux: AuxData
    step_sizes: torch.Tensor
    n_samples: torch.Tensor
    n_accepted: torch.Tensor

    def __init__(
        self,
        logpdfs_aux_fn: Callable[[torch.Tensor], tuple[torch.Tensor, AuxData]],
        init_b: torch.Tensor,
        n_chains: int,
        init_step_size: int | float,
        adapt_rate: int | float,
        target_accept_rate: int | float,
    ):
        """Initializes the Metropolis-Hastings sampler kernel.

        Args:
            logpdfs_aux_fn (Callable[[torch.Tensor], tuple[torch.Tensor, AuxData]]):
                The log pdfs function with auxiliary data.
            init_b (torch.Tensor): Starting b for the chain.
            n_chains (int): The number of parallel chains to spawn.
            init_step_size (int | float): Kernel step in Metropolis Hastings.
            adapt_rate (int | float): Adaptation rate for the step_size.
            target_accept_rate (int | float): Mean acceptance target.
        """
        dtype = get_dtype()

        self.logpdfs_aux_fn = cast(
            Callable[[torch.Tensor], tuple[torch.Tensor, AuxData]],
            torch.no_grad()(logpdfs_aux_fn),
        )
        self.n_chains = n_chains
        self.adapt_rate = adapt_rate
        self.target_accept_rate = target_accept_rate

        # Initialize b
        self.b = init_b.clone()

        # Compute initial log logpdfs
        self.logpdfs, self.aux = self.logpdfs_aux_fn(self.b)

        # Proposal noise initialization
        self._noise = torch.empty_like(self.b)
        self.step_sizes = torch.full((1, self.b.size(-2)), init_step_size, dtype=dtype)

        # Statistics tracking
        self.n_samples = torch.tensor(0, dtype=torch.int64)
        self.n_accepted = torch.zeros(self.b.size(-2), dtype=dtype)

    def step(self):
        """Performs a single kernel step."""
        # Generate proposal noise
        self._noise.normal_()

        # Get the proposal
        proposed_state = self.b + self._noise * self.step_sizes.unsqueeze(-1)
        proposed_logpdfs, proposed_aux = self.logpdfs_aux_fn(proposed_state)
        logpdf_diff = proposed_logpdfs - self.logpdfs

        # Vectorized acceptance decision
        log_uniform = torch.log(torch.rand_like(logpdf_diff))
        accept_mask = log_uniform < logpdf_diff

        torch.where(accept_mask.unsqueeze(-1), proposed_state, self.b, out=self.b)
        torch.where(accept_mask, proposed_logpdfs, self.logpdfs, out=self.logpdfs)

        torch.where(
            accept_mask.unsqueeze(-1), proposed_aux.psi, self.aux.psi, out=self.aux.psi
        )
        torch.where(
            accept_mask, proposed_aux.logliks, self.aux.logliks, out=self.aux.logliks
        )

        # Update statistics
        self.n_samples += 1
        mean_accept_mask = accept_mask.to(get_dtype()).mean(dim=0)
        self.n_accepted += mean_accept_mask

        # Update step sizes
        adaptation = (mean_accept_mask - self.target_accept_rate) * self.adapt_rate
        self.step_sizes *= torch.exp(adaptation)

    @property
    def acceptance_rates(self) -> torch.Tensor:
        """Gets the acceptance_rate.

        Returns:
            torch.Tensor: The means of the acceptance_rates across iterations.
        """
        return self.n_accepted / torch.clamp(self.n_samples, min=1.0)

    @property
    def mean_acceptance_rate(self) -> float:
        """Gets the acceptance_rate mean across all individuals.

        Returns:
            torch.Tensor: The means across iterations and individuals.
        """
        return self.acceptance_rates.mean().item()

    @property
    def mean_step_size(self) -> float:
        """Gets the mean step size.

        Returns:
            float: The mean step size.
        """
        return self.step_sizes.mean().item()

    @property
    def diagnostics(self) -> dict[str, Any]:
        """Gets the summary of the MCMC diagnostics.

        Returns:
            dict[str, Any]: The dict of the diagnostics.
        """
        return {
            "acceptance_rates": self.acceptance_rates.clone(),
            "mean_acceptance_rate": self.mean_acceptance_rate,
            "step_sizes": self.step_sizes.clone(),
            "mean_step_size": self.mean_step_size,
        }



from typing import Any

import torch

from ..typedefs._defs import LOG_TWO_PI
from ..typedefs._params import ModelParams
from ..utils._linalg import get_cholesky_and_log_eigvals


class PriorMixin:
    """Mixin class for prior model computations."""

    def __init__(self, *args: Any, **kwargs: Any):
        """Initializes the prior mixin."""
        super().__init__(*args, **kwargs)

    @staticmethod
    def _prior_logliks(params: ModelParams, b: torch.Tensor) -> torch.Tensor:
        """Computes the prior log likelihoods.

        Args:
            params (ModelParams): The model parameters.
            b (torch.Tensor): The 3D tensor of random effects.

        Returns:
            torch.Tensor: The computed log likelihoods.
        """
        Q_inv_cholesky, Q_nlog_eigvals = get_cholesky_and_log_eigvals(params, "Q")
        Q_quad_form = (b @ Q_inv_cholesky).pow(2).sum(dim=-1)
        Q_norm_factor = (Q_nlog_eigvals - LOG_TWO_PI).sum()

        return 0.5 * (Q_norm_factor - Q_quad_form)



"""Utility functions for the jmstate package."""

from ._dtype import get_dtype, set_dtype
from ._linalg import cov_from_repr, repr_from_cov
from ._surv import build_buckets

__all__ = ["build_buckets", "cov_from_repr", "get_dtype", "repr_from_cov", "set_dtype"]



import warnings
from collections.abc import Mapping
from dataclasses import field
from functools import cached_property
from typing import Any

import torch
from pydantic import ConfigDict, dataclasses
from rich.tree import Tree

from ..utils._checks import (
    check_consistent_size,
    check_inf,
    check_nan,
    check_trajectory_c,
    check_trajectory_empty,
    check_trajectory_sorting,
)
from ..utils._dtype import get_dtype
from ..utils._surv import build_all_buckets
from ..visualization._print import (
    add_c,
    add_psi,
    add_t,
    add_trajectories,
    add_x,
    add_y,
    rich_str,
)
from ._defs import (
    BaseHazardFn,
    IndividualEffectsFn,
    LinkFn,
    RegressionFn,
    Tensor1D,
    Tensor2D,
    Tensor3D,
    TensorCol,
    Trajectory,
)
from ._params import ModelParams


# Dataclasses
@dataclasses.dataclass(config=ConfigDict(arbitrary_types_allowed=True), frozen=True)
class ModelDesign:
    """Class containing model design.

    For all functions, please use broadcasting as much as possible. It is almost always
    possible to broadcast parameters to vectorize efficiently the operations. If you
    copy, beware of a heavy performance hit. If unable, please use vmap.

    Also, note that the function passed to the MCMC sampler will be built using the
    `torch.no_grad()` decorator. If needs be, use `torch.enable_grad()` if one of the
    model design functions always require gradient computation regardless of setting.

    Ensure all functions all well defined on a closed interval and are differentiable
    almost everywhere.

    Attributes:
        individual_effects_fn (IndividualEffectsFn): The individual effects function. It
            must be able to yield 2D or 3D tensors given inputs of `gamma` (population
            parameters), `x` (covariates matrix), and `b` (random effects). Note `b` is
            either 2D or 3D.
        regression_fn (RegressionFn): The regression function. It
            must be able to yield 3D and 4D tensors given 1D or 2D time inputs, as well
            as `psi` input of order 2 or 3. This is not very restrictive, but requires
            to be careful. The last dimension is the dimension of the response variable;
            second last is the repeated measurements; third last is individual based;
            possible fourth last is for parallelization of the MCMC sampler.
        surv (Mapping[tuple[Any, Any], tuple[BaseHazardFn, LinkFn]]): A mapping of
            transition keys that can be typed however you want. The tuple contains a
            base hazard function in log scale, as well as a link function that shares
            the same requirements as `regression_fn`. Base hazard function is expected
            to be pure if caching is enabled, otherwise it will lead to false
            computations.

    Examples:
        >>> def sigmoid(t: torch.Tensor, psi: torch.Tensor):
        >>>     scale, offset, slope = psi.chunk(3, dim=-1)
        >>>     # Fully broadcasted
        >>>     return (scale * torch.sigmoid((t - offset) / slope)).unsqueeze(-1)
        >>> individual_effects_fn = lambda gamma, x, b: gamma + b
        >>> regression_fn = sigmoid
        >>> surv = {("alive", "dead"): (Exponential(1.2), sigmoid)}
    """

    individual_effects_fn: IndividualEffectsFn
    regression_fn: RegressionFn
    surv: Mapping[
        tuple[Any, Any],
        tuple[BaseHazardFn, LinkFn],
    ]

    def __str__(self) -> str:
        """Return a string representation of the model design.

        Returns:
            str: The string representation.
        """
        tree = Tree("ModelDesign")
        tree.add(f"individual_effects_fn: {self.individual_effects_fn}")
        tree.add(f"regression_fn: {self.regression_fn}")
        surv = tree.add("surv:")
        for k, v in self.surv.items():
            surv.add(f"{k[0]} --> {k[1]}: ({v[0]}, {v[1]})")

        return rich_str(tree)


@dataclasses.dataclass(config=ConfigDict(arbitrary_types_allowed=True), frozen=True)
class ModelData:
    r"""Dataclass containing learnable multistate joint model data.

    Not `y` is expected to be a 3D tensor of dimension :math:`(n, m, d)` if there are
    :math:`n` individual, with a maximum number of :math:`m` measurements in
    :math:`\mathbb{R}^d`. Its values should not be all NaNs.

    Bypass checks by activating the `skip_validation` flag.

    Raises:
        ValueError: If the trajectories are not sorted by time.
        ValueError: If the censoring time is lower than the maximum transition time.
        ValueError: If any of the inputs contain inf values.
        ValueError: If any of the inputs contain NaN values except `y`.
        ValueError: If the size is not consistent between inputs.

    Attributes:
        x (Tensor2D | None): The fixed covariates.
        t (Tensor1D | Tensor2D): The measurement times. Either a 1D tensor if the
            times are shared by all individual, or a matrix of individual times.
            Use padding with NaNs when necessary.
        y (Tensor3D): The measurements. A 3D tensor of dimension :math:`(n, m, d)`
            if there are :math:`n` individual, with a maximum number of :math:`m`
            measurements in :math:`\mathbb{R}^d`. Use padding with NaNs when
            necessary.
        trajectories (list[Trajectory]): The list of the individual trajectories.
            A `Trajectory` is a list of tuples containing time and state.
        c (TensorCol): The censoring times as a column vector. They must not
            be less than the trajectory maximum times.
        skip_validation (bool): Whether to skip validatoin or not. Defaults to False.
    """

    x: Tensor2D | None
    t: Tensor1D | Tensor2D
    y: Tensor3D
    trajectories: list[Trajectory]
    c: TensorCol
    skip_validation: bool = field(default=False, repr=False)

    def __str__(self) -> str:
        """Returns a string representation of the data.

        Returns:
            str: The string representation.
        """
        tree = Tree("ModelData")
        add_x(self.x, tree)
        add_t(self.t, tree)
        add_y(self.y, tree)
        add_trajectories(self.trajectories, tree)
        add_c(self.c, tree)

        return rich_str(tree)

    def __post_init__(self):
        """Runs the post init conversions.

        Raises:
            ValueError: If the trajectories are not sorted by time.
            ValueError: If the censoring time is lower than the maximum transition time.
            ValueError: If any of the inputs contain inf values.
            ValueError: If any of the inputs contain NaN values.
            ValueError: If the size is not consistent between inputs.
        """
        if self.skip_validation:
            return

        dtype = get_dtype()

        object.__setattr__(self, "x", None if self.x is None else self.x.to(dtype))
        object.__setattr__(self, "t", self.t.to(dtype))
        object.__setattr__(self, "y", self.y.to(dtype))
        object.__setattr__(self, "c", self.c.to(dtype))

        check_trajectory_empty(self.trajectories)
        check_trajectory_sorting(self.trajectories)
        check_trajectory_c(self.trajectories, self.c)

        last_times = torch.tensor(
            [trajectory[-1][0] for trajectory in self.trajectories], dtype=dtype
        )

        check_inf(
            (
                (self.x, "x"),
                (self.t, "t"),
                (self.y, "y"),
                (self.c, "c"),
                (last_times, "trajectories"),
            )
        )
        check_nan(((self.x, "x"), (self.c, "c")))
        check_consistent_size(
            (
                (self.x, 0, "x"),
                (self.y, 0, "y"),
                (self.c, 0, "c"),
                (self.size, None, "trajectories"),
            )
        )
        check_consistent_size(((self.t, -1, "t"), (self.y, -2, "y")))

        # Check NaNs between t and y
        if ((~self.y.isnan()).any(dim=-1) & self.t.isnan()).any():
            raise ValueError("NaN time values on non NaN y values")

    @cached_property
    def size(self) -> int:
        """Gets the number of individuals.

        Returns:
            int: The number of individuals.
        """
        return len(self.trajectories)

    def effective_size(self, indep_residuals: bool) -> int:
        """Gets the effective size of the dataset, used for BIC.

        Args:
            indep_residuals (bool): Whether or not to assume the independence of the
                residuals.

        Returns:
            int: The effective size.
        """
        nlong = (
            int((~torch.isnan(self.y)).sum())
            if indep_residuals
            else int((~torch.isnan(self.y)).any(dim=-1).sum())
        )

        return nlong + sum(len(trajectory) - 1 for trajectory in self.trajectories)


class CompleteModelData(ModelData):
    """Complete model data class."""

    valid_mask: Tensor3D = field(init=False)
    n_valid: Tensor2D = field(init=False)
    valid_t: Tensor1D | Tensor2D = field(init=False)
    valid_y: Tensor2D | Tensor3D = field(init=False)
    buckets: dict[tuple[Any, Any], tuple[torch.Tensor, ...]] = field(init=False)

    def prepare(self, model_design: ModelDesign, params: ModelParams):
        """Sets the missing representation.

        Raises:
            ValueError: If y and R are not compatible in shape.

        Args:
            model_design (ModelDesign): The design of the model.
            params (ModelParams): The model parameters.
        """
        check_consistent_size(((params.R_repr.dim, None, "R"), (self.y, -1, "y")))

        nan_mask = self.y.isnan()
        valid_mask = ~nan_mask
        self.valid_mask = valid_mask.to(get_dtype())
        self.n_valid = self.valid_mask.sum(dim=-2)
        self.valid_t = self.t.nan_to_num(self.t.nanmean().item())
        self.valid_y = self.y.nan_to_num()
        self.buckets = build_all_buckets(
            self.trajectories, self.c, tuple(model_design.surv.keys())
        )

        if (
            params.R_repr.method == "full"
            and (valid_mask.any(dim=-1) & nan_mask.any(dim=-1)).any()
        ):
            warnings.warn(
                (
                    "R method should not be full when having mixed NaNs as incorrect "
                    "likelihood will be computed"
                ),
                stacklevel=2,
            )


@dataclasses.dataclass(config=ConfigDict(arbitrary_types_allowed=True), frozen=True)
class SampleData:
    """Dataclass for data used in sampling.

    Bypass checks by activating the `skip_validation` flag.

    Raises:
        ValueError: If the trajectories are not sorted by time.
        ValueError: If the censoring time is lower than the maximum transition time.
        ValueError: If any of the inputs contain inf values.
        ValueError: If any of the inputs contain NaN values.
        ValueError: If the size is not consistent between inputs.

    Attributes:
        x (Tensor2D | None): The fixed covariates.
        trajectories (list[Trajectory]): The list of the individual trajectories.
            A `Trajectory` is a list of tuples containing time and state.
        psi (Tensor2D | Tensor3D): The individual parameters. Define it as a matrix with
            the same number of rows as there are `len(trajectories)`. Only use a 3D
            tensor if you fully understand the codebase and the mechanisms. Trajectory
            sampling may only be used with matrices.
        c (TensorCol | None, optional): The censoring times as a column vector. They
            must not be less than the trajectory maximum times. This corresponds to
            the last times of observation of the individuals or prediction current
            times.
        skip_validation (bool): Whether to skip validatoin or not. Defaults to False.
    """

    x: Tensor2D | None
    trajectories: list[Trajectory]
    psi: Tensor2D | Tensor3D
    c: TensorCol | None = None
    skip_validation: bool = field(default=False, repr=False)

    def __str__(self):
        """Returns a string representation of the data.

        Returns:
            str: The string representation.
        """
        tree = Tree("SampleData")
        add_x(self.x, tree)
        add_trajectories(self.trajectories, tree)
        add_psi(self.psi, tree)
        add_c(self.c, tree)

        return rich_str(tree)

    def __post_init__(self):
        """Runs the post init conversions and checks."""
        if self.skip_validation:
            return

        dtype = get_dtype()

        object.__setattr__(self, "x", None if self.x is None else self.x.to(dtype))
        object.__setattr__(self, "psi", self.psi.to(dtype))
        object.__setattr__(self, "c", None if self.c is None else self.c.to(dtype))

        check_trajectory_empty(self.trajectories)
        check_trajectory_sorting(self.trajectories)
        check_trajectory_c(self.trajectories, self.c)

        last_times = torch.tensor(
            [trajectory[-1][0] for trajectory in self.trajectories], dtype=dtype
        )

        check_inf(
            (
                (self.x, "x"),
                (self.psi, "psi"),
                (self.c, "c"),
                (last_times, "trajectories"),
            )
        )
        check_nan(((self.x, "x"), (self.psi, "psi"), (self.c, "c")))
        check_consistent_size(
            (
                (self.x, -2, "x"),
                (self.psi, -2, "psi"),
                (self.c, -2, "c"),
                (self.size, None, "trajectories"),
            )
        )

    @cached_property
    def size(self) -> int:
        """Gets the number of individuals.

        Returns:
            int: The number of individuals.
        """
        return len(self.trajectories)



from collections.abc import Callable

import torch


def is_valid_dtype(dtype: torch.dtype) -> torch.dtype:
    """Checks if the dtype is valid for PyTorch.

    Args:
        dtype (torch.dtype): The dtype to check.

    Raises:
        ValueError: If the dtype is not float32 or float64.

    Returns:
        torch.dtype: The dtype.
    """
    if dtype not in (torch.float32, torch.float64):
        raise ValueError(
            f"expected dtype in {(torch.float32, torch.float64)}, got {dtype}"
        )
    return dtype


def is_ndim(ndim: int) -> Callable[[torch.Tensor], torch.Tensor]:
    """Checks the number of dimension of a tensor.

    Args:
        ndim (int): The expected ndim.

    Returns:
        Callable[[torch.Tensor], torch.Tensor]: The checker function.
    """

    def _is_ndim(t: torch.Tensor) -> torch.Tensor:
        if t.ndim != ndim:
            raise ValueError(f"expected {ndim}D, got {t.ndim}D")
        return t

    return _is_ndim


def is_col(t: torch.Tensor) -> torch.Tensor:
    """Checks if a tensor is a column vector.

    Args:
        t (torch.Tensor): The tensor to test.

    Raises:
        ValueError: If the tensor is not a column vector.

    Returns:
        torch.Tensor: The output tensor.
    """
    if t.size(1) != 1:
        raise ValueError(f"expected column tensor, got  shape {t.shape}")
    return t


def is_non_neg(x: int | float | torch.Tensor) -> int | float | torch.Tensor:
    """Checks if the argument is non-negative.

    Args:
        x (int | float | torch.Tensor): The input number or tensor.

    Raises:
        ValueError: If the argument is not all non-negative.

    Returns:
        int | float | torch.Tensor: The output number or tensor.
    """
    if (torch.as_tensor(x) < 0).any():
        raise ValueError(f"Expected non-negative number/tensor, got {x}")
    return x


def is_strict_pos(x: int | float | torch.Tensor) -> int | float | torch.Tensor:
    """Checks if the argument is strictly positive.

    Args:
        x (int | float | torch.Tensor): The input number or tensor.

    Raises:
        ValueError: If the argument is not all strictly positive.

    Returns:
        int | float | torch.Tensor: The output number or tensor.
    """
    from jmstate.utils._dtype import get_dtype  # noqa: PLC0415

    dtype = get_dtype()
    x_ = torch.as_tensor(x, dtype=dtype)
    if (x_ <= 0).any():
        raise ValueError(
            f"Expected strictly positive number/tensor when cast to {dtype}, got {x}"
        )
    return x


def is_prob(x: int | float) -> int | float:
    """Checks if the number is a probability.

    Args:
        x (int | float): The number to check.

    Raises:
        ValueError: If the number is not 0 < x < 1.

    Returns:
        int | float: The output number.
    """
    from jmstate.utils._dtype import get_dtype  # noqa: PLC0415

    dtype = get_dtype()
    x_ = torch.as_tensor(x, dtype=dtype)
    if not 0 < x_ < 1:
        raise ValueError(
            f"Expected probability in (0, 1) when cast to {dtype}, got {x}"
        )
    return x



import torch
from pydantic import ConfigDict, validate_call
from torch import nn

from ..typedefs._defs import IntNonNegative, LinkFn, RegressionFn
from ..utils._dtype import get_dtype


def linear(t: torch.Tensor, psi: torch.Tensor) -> torch.Tensor:
    r"""Implements the linear regression or link function.

    When reverting to a linear joint model, this gives the mapping:

    .. math::
        h(t, \psi) = \psi,

    where :math:`\psi` are the individual parameters.

    Args:
        t (torch.Tensor): The time points.
        psi (torch.Tensor): The individual effects (parameters).

    Returns:
        torch.Tensor: The computed transformation.
    """
    return psi.unsqueeze(-2).expand(*psi.shape[:-1], t.size(-1), -1)


class Net(nn.Module):
    r"""Implements a neural network.

    This neural network is very flexible, and any nn.Module may be used, in particular
    sequentials. When not knowing which link or regression function to use, try Net.
    You can use derivatives of arbitrary order for the link function using the
    derivatives method.
    If the input layer is in :math:`\mathbb{R}^d`, then one dimension is used for the
    input :math:`t`, and the rest for the individual parameters.

    Attributes:
        net (nn.Module): The neural network module.

    Examples:
        >>> from torch import nn
        >>> net = Net(nn.Sequential(nn.Linear(3, 4), nn.ReLU(), nn.Linear(4, 1)))
        >>> link = net.derivatives((0, 1)) # derivatives of order 0 and 1
    """

    net: nn.Module

    @validate_call(config=ConfigDict(arbitrary_types_allowed=True))
    def __init__(self, net: nn.Module):
        super().__init__()  # type: ignore
        self.net = net.to(get_dtype())
        self.requires_grad_(False)

    def forward(self, t: torch.Tensor, psi: torch.Tensor) -> torch.Tensor:
        """Implements the neural transformation.

        Args:
            t (torch.Tensor): The time points.
            psi (torch.Tensor): The individual effects (parameters).

        Returns:
            torch.Tensor: The computed transformation.
        """
        psi_ext = psi.unsqueeze(-2).expand(*psi.shape[:-1], t.size(-1), -1)
        t_ext = t.unsqueeze(-1).expand(*psi_ext.shape[:-1], 1)
        x = torch.cat([t_ext, psi_ext], dim=-1)
        return self.net(x)

    @validate_call(config=ConfigDict(arbitrary_types_allowed=True))
    def derivatives(self, degs: tuple[IntNonNegative, ...]) -> RegressionFn | LinkFn:
        """Gets a function returning multiple derivatives of the neural network.

        Args:
            degs (tuple[IntNonNegative, ...]): The degrees.

        Returns:
            RegressionFn | LinkFn: A regresion/link function.
        """
        max_deg = max(degs)

        @torch.enable_grad()  # type: ignore
        def _derivatives(t: torch.Tensor, psi: torch.Tensor) -> torch.Tensor:
            psi_ext = psi.unsqueeze(-2).expand(*psi.shape[:-1], t.size(-1), -1)
            t_ext = t.unsqueeze(-1).expand(*psi_ext.shape[:-1], 1).requires_grad_()
            x = torch.cat([t_ext, psi_ext], dim=-1)
            y = self.net(x)

            needs_grad = y.requires_grad
            ones = torch.ones_like(y)
            out_list = [y] if 0 in degs else []
            for i in range(1, max_deg + 1):
                y = torch.autograd.grad(
                    y, t_ext, ones, create_graph=i < max_deg or needs_grad
                )[0]
                if i in degs:
                    out_list.append(y)

            out = torch.cat(out_list, dim=-1)
            return out if needs_grad else out.detach()

        return _derivatives



"""Job functions for the jmstate package."""

from ._computation import ComputeCriteria, ComputeEBEs, ComputeFIM
from ._fitting import Fit, Scheduling
from ._logging import LogParamsHistory, MCMCDiagnostics
from ._prediction import PredictSurvLogps, PredictTrajectories, PredictY, SwitchParams
from ._projection import AdamL1Proximal
from ._stopping import GradStop, NoStop, ParamStop, ValueStop

__all__ = [
    "AdamL1Proximal",
    "ComputeCriteria",
    "ComputeEBEs",
    "ComputeFIM",
    "Fit",
    "GradStop",
    "LogParamsHistory",
    "MCMCDiagnostics",
    "NoStop",
    "ParamStop",
    "PredictSurvLogps",
    "PredictTrajectories",
    "PredictY",
    "Scheduling",
    "SwitchParams",
    "ValueStop",
]



from bisect import bisect_left
from collections.abc import Callable, Sequence
from functools import partial
from typing import Any, cast

import torch
from pydantic import ConfigDict, validate_call
from rich.console import Console, Group
from rich.panel import Panel
from rich.rule import Rule
from rich.table import Table
from rich.text import Text
from rich.tree import Tree
from torch.distributions import Normal
from tqdm import trange

from ..jobs._computation import ComputeCriteria, ComputeFIM
from ..jobs._fitting import Fit
from ..typedefs._data import CompleteModelData, ModelData, ModelDesign, SampleData
from ..typedefs._defaults import DEFAULT_HYPERPARAMETERS, DEFAULT_HYPERPARAMETERS_FIELDS
from ..typedefs._defs import (
    SIGNIFICANCE_CODES,
    SIGNIFICANCE_LEVELS,
    AuxData,
    Info,
    IntNonNegative,
    IntStrictlyPositive,
    Job,
    Metrics,
    NumNonNegative,
    NumProbability,
    Trajectory,
)
from ..typedefs._params import ModelParams
from ..utils._checks import check_consistent_size, check_inf, check_nan
from ..utils._dtype import get_dtype
from ..utils._misc import run_jobs
from ..visualization._print import rich_str
from ._hazard import HazardMixin
from ._longitudinal import LongitudinalMixin
from ._prior import PriorMixin
from ._sampler import MetropolisHastingsSampler


class MultiStateJointModel(PriorMixin, LongitudinalMixin, HazardMixin):
    r"""A class of the nonlinear multistate joint model.

    It features methods to simulate data, fit based on stochastic gradient with any
    `torch.optim.Optimizer` of choice.

    It leverages the Fisher identity and stochastic gradient algorithm coupled
    with a MCMC (Metropolis Hastings) sampler:

    .. math::
        \nabla_\theta \log \mathcal{L}(\theta ; x) = \mathbb{E}_{b \sim p(\cdot \mid x,
        \theta)} \bigl(\nabla_\theta \log \mathcal{L}(\theta ; x, b)\bigr).

    The use of penalization is possible through the attribute `pen`.

    Please note this class encompasses both the linear joint model and the standard
    joint model, but also allows for the modeling of multiple states assuming a semi
    Markov property.

    Attributes:
        model_design (ModelDesign): The model design.
        params_ (ModelParams): The (variable) model parameters.
        pen (Callable[[ModelParams], torch.Tensor] | None): The log likelihood penalty.
        n_quad (int): The number of nodes for the Gauss Legendre quadrature of hazard.
        n_bisect (int): The number of bisection steps for the bisection algorithm.
        cache_limit (int | None): The limit of the cache used in hazard computation,
            greatly reducing memory and CPU usage. None means infinite, 0 means no
            caching.
        data (ModelData | None): The learnable dataset used when the model was fitted.
        metrics_ (Metrics): Metrics object containing information about the model
            and the jobs it executed.
        fit_ (bool): A boolean value set to True when the model has been fitted.
    """

    model_design: ModelDesign
    params_: ModelParams
    pen: Callable[[ModelParams], torch.Tensor] | None
    n_quad: int
    n_bisect: int
    cache_limit: int | None
    data: ModelData | None
    metrics_: Metrics
    fit_: bool

    @validate_call(config=ConfigDict(arbitrary_types_allowed=True))
    def __init__(
        self,
        model_design: ModelDesign,
        init_params: ModelParams,
        *,
        pen: Callable[[ModelParams], torch.Tensor] | None = None,
        n_quad: IntStrictlyPositive = 32,
        n_bisect: IntStrictlyPositive = 32,
        cache_limit: IntNonNegative | None = 256,
    ):
        """Initializes the joint model based on the user defined design.

        Args:
            model_design (ModelDesign): Model design containing modeling information.
            init_params (ModelParams): Initial values for the parameters.
            pen (Callable[[ModelParams], torch.Tensor] | None, optional):
                The penalization function. Defaults to None.
            n_quad (IntStrictlyPositive, optional): The used number of points for
                Gauss-Legendre quadrature. Defaults to 32.
            n_bisect (IntStrictlyPositive, optional): The number of bisection steps
                used in transition sampling. Defaults to 32.
            cache_limit (IntNonNegative | None, optional): The max length of cache.
                Defaults to 256.
        """
        # Store model components
        self.model_design = model_design
        self.params_ = init_params.clone()

        # Store penalization
        self.pen = pen

        # Info of the Mixin Classes
        super().__init__(
            n_quad=n_quad,
            n_bisect=n_bisect,
            cache_limit=cache_limit,
        )

        # Initialize attributes that will be set later
        self.data = None
        self.metrics_ = Metrics()
        self.fit_ = False

    def __str__(self) -> str:
        """Returns a string representation of the model.

        Returns:
            str: The string representation.
        """
        tree = Tree("MultiStateJointModel")
        tree.add(f"model_design: {self.model_design}")
        tree.add(f"params_: {self.params_}")
        tree.add(f"pen: {self.pen}")
        tree.add(f"n_quad: {self.n_quad}")
        tree.add(f"n_bisect: {self.n_bisect}")
        tree.add(f"cache_limit: {self.cache_limit}")
        tree.add(f"data: {self.data}")
        tree.add(f"metrics_: object with attributes {list(vars(self.metrics_).keys())}")
        tree.add(f"fit_: {self.fit_}")

        return rich_str(tree)

    def summary(self, fmt: str = ".3f"):
        """Prints a summary of the model.

        Args:
            fmt (str, optional): The format of the p-values. Defaults to ".3f".

        This function prints the p-values of the parameters as well as values and
        standard error. Also prints the log likelihood, AIC, BIC with lovely colors!
        """
        named_params_list = self.params_.as_named_list
        values = self.params_.as_flat_tensor
        stderrors = self.stderror.as_flat_tensor
        zvalues = torch.abs(values / stderrors)
        pvalues = cast(torch.Tensor, 2 * (1 - Normal(0, 1).cdf(zvalues)))

        table = Table()
        table.add_column("Parameter name", justify="left")
        table.add_column("Value", justify="center")
        table.add_column("Standard Error", justify="center")
        table.add_column("z-value", justify="center")
        table.add_column("p-value", justify="center")
        table.add_column("Significance level", justify="center")

        i = 0
        for name, value in named_params_list:
            for j in range(1, value.numel() + 1):
                code = SIGNIFICANCE_CODES[
                    bisect_left(SIGNIFICANCE_LEVELS, pvalues[i].item())
                ]

                table.add_row(
                    f"{name}[{j}]" if value.numel() > 1 else name,
                    f"{values[i]:{fmt}}",
                    f"{stderrors[i]:{fmt}}",
                    f"{zvalues[i]:{fmt}}",
                    f"{pvalues[i]:{fmt}}",
                    code,
                )
                i += 1

        criteria = Text(
            f"Log-likelihood: {self.loglik:{fmt}}\n"
            f"Penalized negative log-likelihood: {self.nloglik_pen:{fmt}}\n"
            f"AIC: {self.aic:{fmt}}\n"
            f"BIC: {self.bic:{fmt}}",
            style="bold cyan",
        )

        content = Group(table, Rule(style="dim"), criteria, Rule(style="dim"))

        panel = Panel(
            content, title="Model Summary", border_style="green", expand=False
        )

        Console().print(panel)

    def _logpdfs_aux_fn(
        self, params: ModelParams, b: torch.Tensor, data: CompleteModelData
    ) -> tuple[torch.Tensor, AuxData]:
        """Gets the log pdfs with individual effects and log likelihoods.

        Args:
            params (ModelParams): The model parameters.
            b (torch.Tensor): The random effects.
            data (CompleteModelData): Dataset on which likelihood is computed.

        Returns:
           tuple[torch.Tensor, AuxData]: The log pdfs and aux.
        """
        psi = self.model_design.individual_effects_fn(params.gamma, data.x, b)
        logliks = super()._long_logliks(params, psi, data) + super()._hazard_logliks(
            params, psi, data
        )
        logpdfs = logliks + super()._prior_logliks(params, b)

        return (logpdfs, AuxData(psi, logliks))

    def _setup_mcmc(
        self,
        data: CompleteModelData,
        n_chains: int,
        init_step_size: int | float,
        adapt_rate: int | float,
        target_accept_rate: int | float,
    ) -> MetropolisHastingsSampler:
        """Setup the MCMC kernel and hyper-parameters.

        Args:
            data (CompleteModelData): The complete dataset.
            n_chains (int): The number of parallel MCMC chains.
            init_step_size (int | float): Kernel standard error in Metropolis.
            adapt_rate (int | float): Adaptation rate for the step_size.
            target_accept_rate (int | float): Mean acceptance target.

        Returns:
            MetropolisHastingsSampler: The intialized Markov kernel.
        """
        # Initialize random effects
        init_b = torch.zeros(
            n_chains, data.size, self.params_.Q_repr.dim, dtype=get_dtype()
        )

        return MetropolisHastingsSampler(
            partial(self._logpdfs_aux_fn, self.params_, data=data),
            init_b,
            n_chains,
            init_step_size,
            adapt_rate,
            target_accept_rate,
        )

    def _get_hyperparameters(
        self, job_factories: Sequence[Callable[[Info], Job]]
    ) -> dict[str, Any] | None:
        """Gets default hyper-parameters.

        Args:
            job_factories (Sequence[Callable[[Info], Job]]): The job factories.

        Returns:
            dict[str, Any] | None: The default hyper-parameters associated.
        """
        for job_factory in reversed(job_factories):
            key = getattr(job_factory, "cls", None)
            if not (isinstance(key, type) and issubclass(key, Job)):
                continue

            hyperparameters = DEFAULT_HYPERPARAMETERS.get(key)
            if hyperparameters is not None:
                return hyperparameters

        return None

    @validate_call(config=ConfigDict(arbitrary_types_allowed=True))
    def sample_trajectories(
        self,
        sample_data: SampleData,
        c_max: torch.Tensor,
        *,
        max_length: IntStrictlyPositive = 10,
    ) -> list[Trajectory]:
        """Sample trajectories from the joint model.

        The sampling is done usign a bisection algorithm by inversing the log cdf of the
        transitions inside a Gillespie-like algorithm.

        Checks are run only if the `skip_validation` attribute of `sample_date` is not
        set to `True`.

        Args:
            sample_data (SampleData): Prediction data.
            c_max (TensorCol): The maximum trajectory censoring times.
            max_length (IntStrictlyPositive, optional): Maximum iterations or sampling.
                Defaults to 10.

        Raises:
            ValueError: If c_max contains inf values.
            ValueError: If c_max contains NaN values.
            ValueError: If c_max has incorrect shape.

        Returns:
            list[Trajectory]: The sampled trajectories.
        """
        c_max = c_max.to(get_dtype())

        if not sample_data.skip_validation:
            check_inf(((c_max, "c_max"),))
            check_nan(((c_max, "c_max"),))
            check_consistent_size(
                ((c_max, 0, "c_max"), (sample_data.size, None, "sample_data.size"))
            )

        return super()._sample_trajectories(sample_data, c_max, max_length=max_length)

    @validate_call(config=ConfigDict(arbitrary_types_allowed=True))
    def compute_surv_logps(
        self, sample_data: SampleData, u: torch.Tensor
    ) -> torch.Tensor:
        r"""Computes log probabilites of remaining event free up to time u.

        A censoring time may also be given. With known individual effects, this computes
        at the times :math:`u` the values of the log survival probabilities given input
        data conditionally to survival up to time :math:`c`:

        .. math::
            \log \mathbb{P}(T^* \geq u \mid T^* > c) = -\int_c^u \lambda(t) \, dt.

        When multiple transitions are allowed, :math:`\lambda(t)` is a sum over all
        possible transitions, that is to say if an individual is in the state :math:`k`
        from time :math:`t_0`, this gives:

        .. math::
            -\int_c^u \sum_{k'} \lambda^{k' \mid k}(t \mid t_0) \, dt.

        Please note this makes use of the Chasles property in order to avoid the
        computation of two integrals and make computations more precise.

        The variable `u` is expected to be a matrix with the same number of rows as
        individuals, and the same number of columns as prediction times.

        Checks are run only if the `skip_validation` attribute of `sample_date` is not
        set to `True`.

        Args:
            sample_data (SampleData): The data on which to compute the probabilities.
            u (Tensor2D): The time at which to evaluate the probabilities.

        Raises:
            ValueError: If u contains inf values.
            ValueError: If u contains NaN values.
            ValueError: If u has incorrect shape.

        Returns:
            torch.Tensor: The computed survival log probabilities.
        """
        u = u.to(get_dtype())

        if not sample_data.skip_validation:
            check_inf(((u, "u"),))
            check_nan(((u, "u"),))
            check_consistent_size(
                ((u, 0, "u"), (sample_data.size, None, "sample_data.size"))
            )

        return super()._compute_surv_logps(sample_data, u)

    @validate_call(config=ConfigDict(arbitrary_types_allowed=True))
    def do(
        self,
        data: ModelData,
        job_factories: Sequence[Callable[[Info], Job]],
        *,
        max_iterations: int | None = None,
        n_chains: IntStrictlyPositive | None = None,
        warmup: IntNonNegative | None = None,
        n_steps: IntStrictlyPositive | None = None,
        init_step_size: NumNonNegative = 0.1,
        adapt_rate: NumNonNegative = 0.1,
        accept_target: NumProbability = 0.234,
        verbose: bool = True,
    ) -> Metrics:
        """Runs the sampler loop and some jobs.

        Many jobs are predefined for user convenience, but you can use the base class
        `Job` to define your own. The `Job` class returns a factory that is not
        initialized until this function calls the job factories. For default jobs, a set
        of default hyper-parameters allows matching via the `.cls` attribute, but these
        defaults can always be overriden. In the case where multiple jobs with
        conflicting defaults are passed, only the last default will be kept and used.

        To do parallel MCMC sampling, which is enabled by default, use `n_chains`.

        Also, note that the function passed to the MCMC sampler will be built using the
        `torch.no_grad()` decorator. If needs be, use `torch.enable_grad()` if one of
        the model design functions always require gradient computation regardless of
        setting. If the data provided is equal to the data seen during fit, metrics are
        computed and added to those computed during fit in the `metrics_` attribute.

        To enable caching, please refer to the argument `cache_limit` to see the
        behaviour.

        This returns a `Metrics` object.

        Args:
            data (ModelData): The dataset to learn from.
            job_factories (Sequence[Callable[[Info], Job]]): A sequence of job factories
                to execute in the order in which they are given.
            max_iterations (int | None, optional): Maximum number of iterations.
                Defaults to None.
            n_chains (IntStrictlyPositive | None, optional): Batch size used. Defaults
                to None.
            warmup (IntNonNegative | None, optional): The number of iteration steps used
            in the warmup. Defaults to None.
            n_steps (IntStrictlyPositive | None, optional): The steps to do at each
                iteration; this is sub-sampling. Defaults to None.
            init_step_size (NumNonNegative, optional): Initial kernel step size in
                Metropolis Hastings. Defaults to 0.1.
            adapt_rate (NumNonNegative, optional): Adaptation rate for the step_size.
                The adaptation is done with the Robbins Monro algorithm in log scale.
                Defaults to 0.1.
            accept_target (NumProbability, optional): Acceptance target. Defaults to
                0.234.
            verbose (bool, optional): Whether or not to show progress. Defaults to True.

        Raises:
            TypeError: If some attribute is left unset.

        Returns:
            Metrics: The metrics.
        """
        # Load and complete data
        complete_data = CompleteModelData(
            data.x, data.t, data.y, data.trajectories, data.c, skip_validation=True
        )
        complete_data.prepare(self.model_design, self.params_)

        # Set up jobs and hyper-parameters.
        hyperparameters = self._get_hyperparameters(job_factories)
        if hyperparameters is not None:
            max_iterations = (
                hyperparameters["max_iterations"]
                if max_iterations is None
                else max_iterations
            )
            n_chains = hyperparameters["n_chains"] if n_chains is None else n_chains
            warmup = hyperparameters["warmup"] if warmup is None else warmup
            n_steps = hyperparameters["n_steps"] if n_steps is None else n_steps

        # Check everything is there
        for field in DEFAULT_HYPERPARAMETERS_FIELDS:
            if locals()[field] is None:
                raise TypeError(f"Missing required argument: '{field}'")

        # Set up MCMC
        sampler = self._setup_mcmc(
            complete_data,
            cast(int, n_chains),
            init_step_size,
            adapt_rate,
            accept_target,
        )
        for _ in range(cast(int, warmup)):
            sampler.step()

        # Initialize info
        info = Info(
            data=data,
            logpdfs_aux_fn=partial(self._logpdfs_aux_fn, data=complete_data),
            iteration=-1,
            model=self,
            sampler=sampler,
        )

        jobs = [job_factory(info) for job_factory in job_factories]

        # Main loop
        for _ in trange(
            cast(int, max_iterations), desc="Running joint model", disable=not verbose
        ):
            info.iteration += 1
            if run_jobs(jobs, info):
                break

            for _ in range(cast(int, n_steps)):
                sampler.step()

        # End things
        metrics = self.metrics_ if data is self.data else Metrics()

        info.iteration += 1
        for job in jobs:
            job.end(info=info, metrics=metrics)

        self._cache.clear_cache()
        return metrics

    @validate_call(config=ConfigDict(arbitrary_types_allowed=True))
    def fit(
        self,
        data: ModelData,
        job_factories: Sequence[Callable[[Info], Job]] = (),
        *,
        max_iterations: int | None = None,
        n_chains: IntStrictlyPositive | None = None,
        warmup: IntNonNegative | None = None,
        n_steps: IntStrictlyPositive | None = None,
        init_step_size: NumNonNegative = 0.1,
        adapt_rate: NumNonNegative = 0.1,
        accept_target: NumProbability = 0.234,
        verbose: bool = True,
        opt_factory: type[torch.optim.Optimizer] | None = None,
        fit_extra: bool = True,
        **kwargs: Any,
    ) -> Metrics:
        """Fits the model and runs some jobs.

        Interally calls
        `do(data, [Fit(opt_factory, fit_extra, lr, **kwargs), *job_factories], ...)`.
        See `do` for more details.

        This sets the `fit_` attribute to True and the `data` attribute to the data
        provided. Metrics are stored in the `metrics_` attribute.

        Args:
            data (ModelData): The dataset to learn from.
            job_factories (Sequence[Callable[[Info], Fit]], optional): A sequence of job
                factories to execute in the order in which they are given. Defaults to
                ()
            max_iterations (int | None, optional): Maximum number of iterations.
                Defaults to None.
            n_chains (IntStrictlyPositive | None, optional): Batch size used. Defaults
                to None.
            warmup (IntNonNegative | None, optional): The number of iteration steps used
            in the warmup. Defaults to None.
            n_steps (IntStrictlyPositive | None, optional): The steps to do at each
                iteration; this is sub-sampling. Defaults to None.
            init_step_size (NumNonNegative, optional): Initial kernel step size in
                Metropolis Hastings. Defaults to 0.1.
            adapt_rate (NumNonNegative, optional): Adaptation rate for the step_size.
                The adaptation is done with the Robbins Monro algorithm in log scale.
                Defaults to 0.1.
            accept_target (NumProbability, optional): Acceptance target. Defaults to
                0.234.
            verbose (bool, optional): Whether or not to show progress. Defaults to True.
            opt_factory (type[torch.optim.Optimizer] | None, optional): The optimizer
                factory. Defaults to None.
            fit_extra (bool, optional): An option to fit extra parameters or not.
                Defaults to True.
            kwargs (Any): Additional kwargs passed to the optimizer factory, such as lr.

        Returns:
            Metrics: The metrics computed.
        """
        return self.do(
            data,
            [Fit(opt_factory, fit_extra, **kwargs), *job_factories],
            max_iterations=max_iterations,
            n_chains=n_chains,
            warmup=warmup,
            n_steps=n_steps,
            init_step_size=init_step_size,
            adapt_rate=adapt_rate,
            accept_target=accept_target,
            verbose=verbose,
        )

    def compute_summary(
        self,
        job_factories: Sequence[Callable[[Info], Job]] = (),
        *,
        max_iterations: int | None = None,
        n_chains: IntStrictlyPositive | None = None,
        warmup: IntNonNegative | None = None,
        n_steps: IntStrictlyPositive | None = None,
        init_step_size: NumNonNegative = 0.1,
        adapt_rate: NumNonNegative = 0.1,
        accept_target: NumProbability = 0.234,
        verbose: bool = True,
        bias: bool = True,
    ) -> Metrics:
        """Fits the model and runs some jobs.

        Interally calls
        `do(self.data, [ComputeFIM(bias), ComputeCriteria(), *job_factories], ...)`. See
        `do` for more details.

        Args:
            job_factories (Sequence[Callable[[Info], Job]], optional): A sequence of job
                factories to execute in the order in which they are given.
            max_iterations (int | None, optional): Maximum number of iterations.
                Defaults to None.
            n_chains (IntStrictlyPositive | None, optional): Batch size used. Defaults
                to None.
            warmup (IntNonNegative | None, optional): The number of iteration steps used
            in the warmup. Defaults to None.
            n_steps (IntStrictlyPositive | None, optional): The steps to do at each
                iteration; this is sub-sampling. Defaults to None.
            init_step_size (NumNonNegative, optional): Initial kernel step size in
                Metropolis Hastings. Defaults to 0.1.
            adapt_rate (NumNonNegative, optional): Adaptation rate for the step_size.
                The adaptation is done with the Robbins Monro algorithm in log scale.
                Defaults to 0.1.
            accept_target (NumProbability, optional): Acceptance target. Defaults to
                0.234.
            verbose (bool, optional): Whether or not to show progress. Defaults to True.
            bias (bool, optional): Whether or not to substract the bias term of the
                gradient in the FIM. Refer to `ComputeFIM` for more details. Defaults to
                True.

        Raises:
            ValueError: If self.data is None.

        Returns:
            Metrics: The metrics computed.
        """
        if self.data is None:
            raise ValueError("Data must be set before and seen during fitting")

        return self.do(
            self.data,
            [ComputeFIM(bias), ComputeCriteria(), *job_factories],
            max_iterations=max_iterations,
            n_chains=n_chains,
            warmup=warmup,
            n_steps=n_steps,
            init_step_size=init_step_size,
            adapt_rate=adapt_rate,
            accept_target=accept_target,
            verbose=verbose,
        )

    @validate_call(config=ConfigDict(arbitrary_types_allowed=True))
    def sample_params(self, sample_size: IntNonNegative) -> list[ModelParams]:
        """Sample parameters based on asymptotic behavior of the MLE.

        Args:
            sample_size (IntNonNegative): The desired sample size.

        Returns:
            list[ModelParams]: A list of model parameters.
        """
        dist = torch.distributions.MultivariateNormal(
            self.params_.as_flat_tensor, self.fim.inverse()
        )
        flat_samples = dist.sample((sample_size,))

        return [self.params_.from_flat_tensor(sample) for sample in flat_samples]

    @property
    def fim(self) -> torch.Tensor:
        """Returns the Fisher Information Matrix.

        Raises:
            ValueError: If Fisher Information Matrix has not yet been computed.

        Returns:
            torch.Tensor: The Fisher Information Matrix.
        """
        if not hasattr(self.metrics_, "fim"):
            raise ValueError("Fisher Information Matrix must be previously computed.")

        return self.metrics_.fim

    @property
    def loglik(self) -> float:
        """Returns the log likelihood of the model.

        Returns:
            float: The log likelihood.
        """
        if not hasattr(self.metrics_, "loglik"):
            raise ValueError("Log likelihood must be previously computed.")

        return self.metrics_.loglik

    @property
    def nloglik_pen(self) -> float:
        """Returns the penalized negative log likelihood of the model.

        Returns:
            float: The penalized negative log likelihood.
        """
        if not hasattr(self.metrics_, "nloglik_pen"):
            raise ValueError(
                "Penalized negative log likelihood must be previously computed."
            )

        return self.metrics_.nloglik_pen

    @property
    def aic(self) -> float:
        """Returns the Akaike Information Criterion of the model.

        Returns:
            float: The Akaike Information Criterion.
        """
        if not hasattr(self.metrics_, "aic"):
            raise ValueError("AIC must be previously computed.")

        return self.metrics_.aic

    @property
    def bic(self) -> float:
        """Returns the Bayesian Information Criterion of the model.

        Returns:
            float: The Bayesian Information Criterion.
        """
        if not hasattr(self.metrics_, "bic"):
            raise ValueError("BIC must be previously computed.")

        return self.metrics_.bic

    @property
    def stderror(self) -> ModelParams:
        r"""Returns the standard error of the parameters.

        They can be used to draw confidence intervals. The standard errors are computed
        using the diagonal of the inverse of the inverse Fisher Information Matrix at
        the MLE:

        .. math::
            \text{sd} = \sqrt{\operatorname{diag}\bigl(\mathcal{I}(\hat{\theta})^{-1}
            \bigr)}

        Returns:
            ModelParams: The standard error in the same format as the parameters.
        """
        return self.params_.from_flat_tensor(self.fim.inverse().diag().sqrt())



from __future__ import annotations

from typing import TYPE_CHECKING, cast

import torch
from pydantic import ConfigDict, validate_call

from ..typedefs._defs import MatRepr, Tensor2D

if TYPE_CHECKING:
    from ..typedefs._params import ModelParams


def _tril_from_flat(flat: torch.Tensor, dim: int) -> torch.Tensor:
    """Generate the lower triangular matrix associated with flat tensor.

    Args:
        flat (torch.Tensor): Flat tensor
        dim (int): Dimension of the matrix.

    Returns:
        torch.Tensor: The lower triangular matrix.
    """
    return torch.zeros(dim, dim, dtype=flat.dtype).index_put_(
        tuple(torch.tril_indices(dim, dim)), flat
    )


def _flat_from_tril(L: torch.Tensor) -> torch.Tensor:
    """Flatten the lower triangular part (including the diagonal) of a square matrix.

    Into a 1D tensor, in row-wise order.

    Args:
        L (torch.Tensor): Square lower-triangular matrix of shape (dim, dim).

    Raises:
        RuntimeError: If the flattening fails.

    Returns:
        torch.Tensor: Flattened 1D tensor containing the lower triangular entries.
    """
    dim = L.size(0)
    return L[tuple(torch.tril_indices(dim, dim))]


def _log_cholesky_from_flat(
    flat: torch.Tensor, dim: int, method: str = "full"
) -> torch.Tensor:
    """Computes log cholesky from flat tensor according to choice of method.

    Args:
        flat (torch.Tensor): The flat tensor parameter.
        dim (int): The dimension of the matrix.
        method (str, optional): The method, full, diag or ball. Defaults to "full".

    Raises:
        ValueError: If the method is not in ("full", "diag", "ball").

    Returns:
        torch.Tensor: The log cholesky representation.
    """
    match method:
        case "full":
            return _tril_from_flat(flat, dim)
        case "diag":
            return torch.diag(flat)
        case "ball":
            return flat * torch.eye(dim)
        case _:
            raise ValueError(
                f"Method must be be either 'full', 'diag' or 'ball', got {method}"
            )


def _flat_from_log_cholesky(L: torch.Tensor, method: str = "full") -> torch.Tensor:
    """Computes flat tensor from log cholesky matrix according to choice of method.

    Args:
        L (torch.Tensor): The square lower triangular matrix parameter.
        method (str, optional): The method, full, diag or ball. Defaults to "full".

    Raises:
        ValueError: If the method is not in ("full", "diag", "ball").

    Returns:
        torch.Tensor: The flat representation.
    """
    match method:
        case "full":
            return _flat_from_tril(L)
        case "diag":
            return L.diag()
        case "ball":
            return L[0, 0].reshape(1)
        case _:
            raise ValueError(
                f"Method must be be either 'full', 'diag' or 'ball', got {method}"
            )


@validate_call(config=ConfigDict(arbitrary_types_allowed=True))
def cov_from_repr(mat_repr: MatRepr) -> Tensor2D:
    r"""Computes covariance matrix from representation.

    Note three types of covariance matrices parametrization are provided: scalar
    matrix; diagonal matrix; full matrix. Defaults to the full matrix parametrization.
    This is achieved through a log Cholesky parametrization of the inverse covariance
    matrix. Formally, consider :math:`P = \Sigma^{-1}` the precision matrix and let
    :math:`L` be the Cholesky factor with positive diagonal elements, the log Cholseky
    is given by:

    .. math::
        \tilde{L}_{ij} = L_{ij}, \, i > j,

    and:

    .. math::
        \tilde{L}_{ii} = \log L_{ii}.

    This is very numerically stable and fast, as it doesn't require inverting the
    matrix when computing quadratic forms. The log determinant is then equal to:

    .. math::

        \log \det P = 2 \operatorname{Tr}(\tilde{L}).

    You can use these methods by creating the appropriate `MatRepr` with methods of
    `ball`, `diag` or `full`.

    Args:
        mat_repr (MatRepr): The matrix representation.

    Raises:
        ValueError: If method 'full' and number of elements are inconsistent with dim.
        ValueError: If method 'diag' and number of elements are inconsistent with dim.
        ValueError: If method 'ball' and number of elements is not one.

    Returns:
        torch.Tensor: The covariance matrix.
    """
    flat, dim, method = mat_repr

    if method == "full" and flat.numel() != (dim * (dim + 1)) // 2:
        raise ValueError(
            f"Inconsistent dim:{dim} with method 'full', flat with {flat.numel()} "
            "elements"
        )
    if method == "diag" and flat.numel() != dim:
        raise ValueError(
            f"Inconsistent dim:{dim} with method 'diag', flat with {flat.numel()} "
            "elements"
        )
    if method == "ball" and flat.numel() != 1:
        raise ValueError("Inconsistent with method 'ball', flat must have one element")

    L = _log_cholesky_from_flat(flat, dim, method)
    L.diagonal().exp_()

    return torch.cholesky_inverse(L)


@validate_call(config=ConfigDict(arbitrary_types_allowed=True))
def repr_from_cov(V: Tensor2D, method: str = "full") -> MatRepr:
    r"""Computes representation from covariance matrix according to choice of method.

    Note three types of covariance matrices parametrization are provided: scalar
    matrix; diagonal matrix; full matrix. Defaults to the full matrix parametrization.
    This is achieved through a log Cholesky parametrization of the inverse covariance
    matrix. Formally, consider :math:`P = \Sigma^{-1}` the precision matrix and let
    :math:`L` be the Cholesky factor with positive diagonal elements, the log Cholseky
    is given by:

    .. math::
        \tilde{L}_{ij} = L_{ij}, \, i > j,

    and:

    .. math::
        \tilde{L}_{ii} = \log L_{ii}.

    This is very numerically stable and fast, as it doesn't require inverting the
    matrix when computing quadratic forms. The log determinant is then equal to:

    .. math::

        \log \det P = 2 \operatorname{Tr}(\tilde{L}).

    You can use these methods by creating the appropriate `MatRepr` with methods of
    `ball`, `diag` or `full`.

    Args:
        V (Tensor2D): The square covariance matrix parameter.
        method (str, optional): The method, full, diag or ball. Defaults to "full".

    Returns:
        MatRepr: The flat representation.
    """
    L = cast(torch.Tensor, torch.linalg.cholesky(V.inverse()))  # type: ignore
    L.diagonal().log_()
    return MatRepr(_flat_from_log_cholesky(L, method), L.size(0), method)


def get_cholesky_and_log_eigvals(
    params: ModelParams, matrix: str
) -> tuple[torch.Tensor, torch.Tensor]:
    """Gets Cholesky factor as well as log eigvals.

    Args:
        params (ModelParams): The model parameters.
        matrix (str): Either "Q" or "R".

    Returns:
        tuple[torch.Tensor, torch.Tensor]: Precision matrix and log eigvals.
    """
    # Get flat then log cholesky
    flat, dim, method = getattr(params, matrix + "_repr")

    L = _log_cholesky_from_flat(flat, dim, method)
    log_eigvals = 2 * L.diag()
    L.diagonal().exp_()

    return L, log_eigvals



"""multistate joint modeling package.

This package provides tools for multistate joint modeling with PyTorch.
"""

from .model import MultiStateJointModel

__author__ = "Félix Laplante"
__email__ = "felixlaplante0@gmail.com"
__license__ = "MIT"

__all__ = ["MultiStateJointModel"]



from collections.abc import Callable
from typing import Any

from ..typedefs._defs import Info, Job, Metrics
from ..typedefs._params import ModelParams


class LogParamsHistory(Job):
    """Job to log the evolution of the paramters during fit.

    This yields a list of `ModelParams`.

    Attributes:
        params_history (list[ModelParams]): The history of the model parameters.
    """

    params_history: list[ModelParams]

    def __new__(cls) -> Callable[[Info], Job]:
        """Creates the parameter logging job."""
        return super().__new__(cls)

    def __init__(self, **_kwargs: Any):  # type: ignore
        """Initializes the history to an empty list."""
        self.params_history = []

    def run(self, info: Info):
        """Appends the current parameter values to the parameter list.

        Args:
            info (Info): The job information object.
        """
        self.params_history.append(info.model.params_.clone().detach())

    def end(self, metrics: Metrics, **_kwargs: Any):
        """Adds by concatenation or erases the metrics history.

        Args:
            metrics (Metrics): The metrics object.
        """
        if not hasattr(metrics, "params_history"):
            metrics.params_history = self.params_history
        else:
            metrics.params_history += self.params_history


class MCMCDiagnostics(Job):
    """Job to log the evolution of the MCMC sampler.

    This yields a list of dicts containing MCMC diagnostics, such as step sizes and
    acceptance rates.

    Attributes:
        mcmc_diagnostics (list[dict[str, Any]]): A list of MCMC diagnostic information.
    """

    mcmc_diagnostics: list[dict[str, Any]]

    def __new__(cls) -> Callable[[Info], Job]:
        """Creates the MCMC diagnostics logging job."""
        return super().__new__(cls)

    def __init__(self, **_kwargs: Any):  # type: ignore
        """Initializes the diagnostics to an empty list."""
        self.mcmc_diagnostics = []

    def run(self, info: Info):
        """Appends the current diagnostics dict to the parameter list.

        Args:
            info (Info): The job information object.
        """
        self.mcmc_diagnostics.append(info.sampler.diagnostics)

    def end(self, metrics: Metrics, **_kwargs: Any):
        """Adds by concatenation or erases the metrics MCMC diagnostics.

        Args:
            metrics (Metrics): The metrics object.
        """
        if not hasattr(metrics, "mcmc_diagnostics"):
            metrics.mcmc_diagnostics = self.mcmc_diagnostics
        else:
            metrics.mcmc_diagnostics += self.mcmc_diagnostics



import torch
from pydantic import ConfigDict, validate_call

from ..typedefs._defs import ValidDtype

# Default dtype
_dtype = torch.float32


@validate_call(config=ConfigDict(arbitrary_types_allowed=True))
def set_dtype(dtype: ValidDtype):
    """Set the default dtype. It must be either `torch.float32` or `torch.float64`.

    It defaults to `torch.float32`.

    Args:
        dtype (ValidDtype): The dtype to set.
    """
    global _dtype  # noqa: PLW0603
    _dtype = dtype


def get_dtype() -> torch.dtype:
    """Get the default dtype. It is either `torch.float32` or `torch.float64`.

    It defaults to `torch.float32`.

    Returns:
        torch.dtype: The default dtype.
    """
    return _dtype



import warnings
from collections.abc import Callable
from typing import Any, Final

import torch
from pydantic import ConfigDict, validate_call

from ..typedefs._defs import Info, Job

# Constants
DEFAULT_OPT_FACTORY: Final[type[torch.optim.Optimizer]] = torch.optim.Adam
DEFAULT_OPT_KWARGS: Final[dict[str, Any]] = {"lr": 0.5}


class Fit(Job):
    """Job to fit the model with random effects.

    This is a fitting job. It sets the `fit_` attribute when finished.

    It can be used with any optimizer factory built on the base class
    `torch.optim.Optimizer`. If None, it defaults to Adam.

    Use kwargs to pass defaults to the optimizer factory.

    Change the value of `fit_extra` if you do not want to fit extra parameters.

    It warns the user if the parameters contain infinite or NaN values.

    Attributes:
        fit_extra (bool): Whether to set the model `fit` attribute or not.
        opt (torch.optim.Optimizer): The optimizer object.

    Examples:
        >>> Fit(torch.optim.Adam, lr=0.01)
    """

    fit_extra: bool
    opt: torch.optim.Optimizer

    @validate_call(config=ConfigDict(arbitrary_types_allowed=True))
    def __new__(
        cls,
        opt_factory: type[torch.optim.Optimizer] | None = None,
        fit_extra: bool = True,
        **kwargs: Any,
    ) -> Callable[[Info], Job]:
        """Creates the fitting class. By default, it uses Adam and lr=0.5.

        Args:
            opt_factory (type[torch.optim.Optimizer] | None, optional): The optimizer
                factory. Defaults to None.
            fit_extra (bool, optional): An option to fit extra parameters or not.
                Defaults to True.
            kwargs (Any): Additional kwargs passed to the optimizer factory, such as lr.
        """
        return super().__new__(cls, opt_factory, fit_extra, **kwargs)

    def __init__(  # type: ignore
        self,
        opt_factory: type[torch.optim.Optimizer] | None = None,
        fit_extra: bool = True,
        *,
        info: Info,
        **kwargs: Any,
    ):
        """Initializes the fitting class. By default, it uses Adam and lr=0.5.

        Args:
            opt_factory (type[torch.optim.Optimizer] | None): The optimizer factory.
                Defaults to None.
            fit_extra (bool, optional): An option to fit extra parameters or not.
                Defaults to True.
            info (Info): The job information object.
            kwargs (Any): Additional kwargs passed to the optimizer factory, such as lr.
        """
        info.model.data = info.data
        self.fit_extra = fit_extra

        info.model.params_.requires_grad_(True)
        if self.fit_extra:
            info.model.params_.extra_requires_grad_(True)

        opt_params = info.model.params_.as_list + (info.model.params_.extra or [])
        self.opt = (opt_factory or DEFAULT_OPT_FACTORY)(
            opt_params, **{**DEFAULT_OPT_KWARGS, **kwargs}
        )
        info.opt = self.opt

    def run(self, info: Info):
        """Performs a single optimization step using the optimizer.

        This calls the optimizer's `step` function, which in turn calls the `closure`
        method to compute the loss and gradients.

        Args:
            info (Info): The job information object.
        """

        def _closure():
            self.opt.zero_grad()  # type: ignore
            logpdfs, _ = info.logpdfs_aux_fn(info.model.params_, info.sampler.b)
            loss = (
                -logpdfs.mean()
                if info.model.pen is None
                else -logpdfs.mean() + info.model.pen(info.model.params_)
            )
            loss.backward()  # type: ignore
            return loss.detach()

        self.opt.step(_closure)  # type: ignore

        info.sampler.logpdfs, info.sampler.aux = info.sampler.logpdfs_aux_fn(
            info.sampler.b
        )

    def end(self, info: Info, **_kwargs: Any):
        """Ends fitting cycle.

        This resets the parameters gradient and sets the `fit_` model attribute to
        True.

        Args:
            info (Info): The job information object.
        """
        params_flat_tensor = info.model.params_.as_flat_tensor
        if (
            torch.isnan(params_flat_tensor).any()
            or torch.isinf(params_flat_tensor).any()
        ):
            warnings.warn("Error infering model parameters", stacklevel=2)

        info.model.params_.requires_grad_(False)
        if self.fit_extra:
            info.model.params_.extra_requires_grad_(False)

        info.model.fit_ = True


class Scheduling(Job):
    """Job to run a scheduler during the optimization.

    It can be used with any optimizer factory built on the base class
    `torch.optim.lr_scheduler.LRScheduler`.

    Use kwargs to pass defaults to the scheduler factory.

    Attributes:
        sched (torch.optim.lr_scheduler.LRScheduler): The scheduler object.
    """

    sched: torch.optim.lr_scheduler.LRScheduler

    @validate_call(config=ConfigDict(arbitrary_types_allowed=True))
    def __new__(
        cls,
        sched_factory: type[torch.optim.lr_scheduler.LRScheduler],
        **kwargs: Any,
    ) -> Callable[[Info], Job]:
        """Creates the scheduler.

        Args:
            sched_factory (type[torch.optim.lr_scheduler.LRScheduler]): The scheduler
                factory.
            kwargs (Any): Additional kwargs passed to the scheduler factory.
        """
        return super().__new__(cls, sched_factory, **kwargs)

    def __init__(
        self,
        sched_factory: type[torch.optim.lr_scheduler.LRScheduler],
        *,
        info: Info,
        **kwargs: Any,
    ):
        """Initializes the scheduler.

        Raises:
            ValueError: If the optimizer has not been initialized before the scheduler.

        Args:
            sched_factory (type[torch.optim.lr_scheduler.LRScheduler]): The scheduler
                factory.
            info (Info): The job information object.
            kwargs (Any): Additional kwargs passed to the scheduler factory.
        """
        if not hasattr(info, "opt"):
            raise ValueError("Optimizer must be initialized before scheduler")

        self.sched = sched_factory(info.opt, **kwargs)

    def run(self, **_kwargs: Any):
        """Performs a single scheduling step using the optimizer."""
        self.sched.step()



from __future__ import annotations

from collections.abc import Callable
from typing import TYPE_CHECKING, Any, cast

import numpy as np
import torch

from ..typedefs._defs import Info, Job
from ..utils._dtype import get_dtype

if TYPE_CHECKING:
    from ..typedefs._params import ModelParams


def legendre_quad(n_quad: int) -> tuple[torch.Tensor, torch.Tensor]:
    """Get the Legendre quadrature nodes and weights.

    Args:
        n_quad (int): The number of quadrature points.

    Returns:
        tuple[torch.Tensor, torch.Tensor]: The nodes and weights.
    """
    nodes, weights = cast(
        tuple[
            np.ndarray[Any, np.dtype[np.float64]],
            np.ndarray[Any, np.dtype[np.float64]],
        ],
        np.polynomial.legendre.leggauss(n_quad),  # type: ignore
    )

    dtype = get_dtype()
    std_nodes = torch.tensor(nodes, dtype=dtype).unsqueeze(0)
    std_weights = torch.tensor(weights, dtype=dtype)

    return std_nodes, std_weights


def map_fn_params(
    params: ModelParams, fn: Callable[[torch.Tensor], torch.Tensor]
) -> ModelParams:
    """Map operation and get new parameters.

    Args:
        params (ModelParams): The model parameters to use.
        fn (Callable[[torch.Tensor], torch.Tensor]): The operation.

    Returns:
        ModelParams: The new parameters (it might be a reshape).
    """
    from ..typedefs._params import ModelParams  # noqa: PLC0415

    def _map_fn(dict: dict[tuple[Any, Any], torch.Tensor]):
        return {key: fn(val) for key, val in dict.items()}

    return ModelParams(
        None if params.gamma is None else fn(params.gamma),
        params.Q_repr._replace(flat=fn(params.Q_repr.flat)),
        params.R_repr._replace(flat=fn(params.R_repr.flat)),
        _map_fn(params.alphas),
        None if params.betas is None else _map_fn(params.betas),
        extra=params.extra,
        skip_validation=True,
    )


def run_jobs(jobs: list[Job], info: Info) -> bool:
    """Call jobs.

    Args:
        jobs (list[Job]): The jobs to execute.
        info (Info): The information container.

    Returns:
        bool: Set to true to stop the iterations.
    """
    stop = None
    for job in jobs:
        result = job.run(info=info)
        stop = (
            stop if result is None else (result if stop is None else (stop and result))
        )

    return False if stop is None else stop



import torch
from rich.console import Console
from rich.tree import Tree

from ..typedefs._defs import Any, Trajectory
from ..utils._surv import build_buckets


# Utils
def rich_str(obj: Any) -> str:
    """Get the rich string representation of an object.

    Args:
        obj (Any): The object to get the string representation of.

    Returns:
        str: The string representation.
    """
    console = Console()
    return console._render_buffer(console.render(obj))[:-1]  # type: ignore


# Descriptors
def add_x(x: torch.Tensor | None, tree: Tree):
    """Add the fixed covariates to the tree.

    Args:
        x (torch.Tensor | None): The fixed covariates.
        tree (Tree): The tree to add to.
    """
    if x is not None:
        tree.add(f"x: {x.size(0)} individual(s) x {x.size(1)} covariate(s)")


def add_t(t: torch.Tensor, tree: Tree):
    """Add the measurement times.

    Args:
        t (torch.Tensor): The measurement times.
        tree (Tree): The tree to add to.
    """
    if t.ndim == 1:
        tree.add(f"t: {t.size(0)} shared measurement(s)")
    else:
        tree.add(f"t: {t.size(0)} individual(s) x {t.size(1)} measurement(s)")


def add_y(y: torch.Tensor, tree: Tree):
    """Add the measurements.

    Args:
        y (torch.Tensor): The measurements.
        tree (Tree): The tree to add to.
    """
    tree.add(
        f"y: {y.size(0)} individual(s) x {y.size(1)} measurement(s) x {y.size(2)} "
        "dimension(s)"
    )


def add_psi(psi: torch.Tensor, tree: Tree):
    """Add the individual parameters.

    Args:
        psi (torch.Tensor): The individual parameters.
        tree (Tree): The tree to add to.
    """
    if psi.ndim == 2:  # noqa: PLR2004
        tree.add(f"psi: {psi.size(0)} individual(s) x {psi.size(1)} parameter(s)")
    else:
        tree.add(
            f"psi: {psi.size(0)} sample(s) x {psi.size(1)} individual(s) x "
            f"{psi.size(2)} parameter(s)"
        )


def add_trajectories(trajectories: list[Trajectory], tree: Tree):
    """Add the trajectories.

    Args:
        trajectories (list[Trajectory]): The trajectories.
        tree (Tree): The tree to add to.
    """
    buckets = build_buckets(trajectories)

    node = tree.add(
        f"trajectories: {len(trajectories)} individual(s) with "
        f"{sum(len(t) - 1 for t in trajectories)} observed transitions"
    )
    for k, v in buckets.items():
        node.add(f"{k[0]} --> {k[1]}: {v.idxs.numel()}")


def add_c(c: torch.Tensor | None, tree: Tree):
    """Add the censoring times.

    Args:
        c (torch.Tensor | None): The censoring times.
        tree (Tree): The tree to add to.
    """
    if c is not None:
        tree.add(f"c: {c.size(0)} censoring time(s)")



"""Type definitions for the jmstate package."""

from ._data import ModelData, ModelDesign, SampleData
from ._defs import (
    BaseHazardFn,
    BucketData,
    ClockMethod,
    IndividualEffectsFn,
    Info,
    Job,
    LinkFn,
    MatRepr,
    Metrics,
    RegressionFn,
)
from ._params import ModelParams

__all__ = [
    "BaseHazardFn",
    "BucketData",
    "ClockMethod",
    "IndividualEffectsFn",
    "Info",
    "Job",
    "LinkFn",
    "MatRepr",
    "Metrics",
    "ModelData",
    "ModelDesign",
    "ModelParams",
    "RegressionFn",
    "SampleData",
]



from dataclasses import field
from functools import cached_property
from itertools import chain
from typing import Any, Self, cast

import torch
from numpy import array2string
from pydantic import ConfigDict, dataclasses, validate_call
from rich.tree import Tree

from ..utils._checks import check_inf, check_matrix_dim, check_nan
from ..utils._dtype import get_dtype
from ..utils._linalg import cov_from_repr
from ..utils._misc import map_fn_params
from ..visualization._print import rich_str
from ._defs import MatRepr, Tensor1D


@dataclasses.dataclass(config=ConfigDict(arbitrary_types_allowed=True), frozen=True)
class ModelParams:
    r"""Dataclass containing model parameters.

    Note three types of covariance matrices parametrization are provided: scalar
    matrix; diagonal matrix; full matrix. Defaults to the full matrix parametrization.
    This is achieved through a log Cholesky parametrization of the inverse covariance
    matrix. Formally, consider :math:`P = \Sigma^{-1}` the precision matrix and let
    :math:`L` be the Cholesky factor with positive diagonal elements, the log Cholseky
    is given by:

    .. math::
        \tilde{L}_{ij} = L_{ij}, \, i > j,

    and:

    .. math::
        \tilde{L}_{ii} = \log L_{ii}.

    This is very numerically stable and fast, as it doesn't require inverting the
    matrix when computing quadratic forms. The log determinant is then equal to:

    .. math::

        \log \det P = 2 \operatorname{Tr}(\tilde{L}).

    You can use these methods by creating the appropriate `MatRepr` with methods of
    `ball`, `diag` or `full`.

    Additionnally, if your data has mixed missing values, do not use `full` matrix
    parametrization for the residuals, as is this case the components must be
    independent.

    Bypass checks by activating the `skip_validation` flag.

    Attributes:
        gamma (Tensor1D | None): The population level parameters.
        Q_repr (MatRepr): The random effects precision matrix representation.
        R_repr (MatRepr): The residuals precision matrix representation.
        alphas (dict[tuple[Any, Any], Tensor1D]) The link linear parameters.
        betas (dict[tuple[Any, Any], Tensor1D] | None): The covariates parameters.
        extra (list[torch.Tensor] | None): A list of parameters that is passed in
            addition to other mandatory parameters.
        skip_validation (bool): A boolean value to skip validation.
    """

    gamma: Tensor1D | None
    Q_repr: MatRepr
    R_repr: MatRepr
    alphas: dict[tuple[Any, Any], Tensor1D]
    betas: dict[tuple[Any, Any], Tensor1D] | None
    extra: list[torch.Tensor] | None = field(default=None, repr=False)
    skip_validation: bool = field(default=False, repr=False)

    def __str__(self) -> str:
        """Return a string representation of the model parameters.

        Returns:
            str: The string representation.
        """

        def _to_str(t: torch.Tensor) -> str:
            return array2string(t.numpy(), precision=3, suppress_small=True)

        tree = Tree("ModelParams")
        if self.gamma is not None:
            tree.add(f"gamma: {_to_str(self.gamma)}")
        tree.add(f"Q: {_to_str(self.Q_repr.flat)}")
        tree.add(f"R: {_to_str(self.R_repr.flat)}")
        alphas = tree.add("alphas:")
        for k, v in self.alphas.items():
            alphas.add(f"{k[0]} --> {k[1]}: {_to_str(v)}")
        if self.betas is not None:
            betas = tree.add("betas:")
            for k, v in self.betas.items():
                betas.add(f"{k[0]} --> {k[1]}: {_to_str(v)}")

        return rich_str(tree)

    def __post_init__(self):
        """Validate and put to dtype all tensors.

        Raises:
            ValueError: If any of the tensors contains inf values.
            ValueError: If any of the tensors contains NaN values.
        """
        if self.skip_validation:
            return

        def _sort_dict(
            dct: dict[tuple[Any, Any], torch.Tensor],
        ) -> dict[tuple[Any, Any], torch.Tensor]:
            return dict(sorted(dct.items(), key=lambda kv: str(kv[0])))

        # For reproducibility purposes and order
        object.__setattr__(self, "alphas", _sort_dict(self.alphas))
        if self.betas is not None:
            object.__setattr__(self, "betas", _sort_dict(self.betas))

        dtype = get_dtype()

        for t in self.as_list:
            t.data = t.to(dtype)
        if self.extra is not None:
            for t in self.extra:
                t.data = t.to(dtype)

        check_matrix_dim(self.Q_repr, "Q")
        check_matrix_dim(self.R_repr, "R")
        for key, val in self.as_dict.items():
            if isinstance(val, dict):
                tensor_tuple = tuple((t, key) for t in val.values())
                check_inf(tensor_tuple)
                check_nan(tensor_tuple)
            else:
                check_inf(((val, key),))
                check_nan(((val, key),))

    @cached_property
    def as_dict(self) -> dict[str, torch.Tensor | dict[tuple[Any, Any], torch.Tensor]]:
        """Gets a grouped dict of all the parameters.

        Returns:
            dict[str, torch.Tensor | dict[tuple[Any, Any], torch.Tensor]]: The dict of
                the parameters.
        """
        groups = {
            "gamma": self.gamma,
            "Q": self.Q_repr.flat,
            "R": self.R_repr.flat,
            "alphas": self.alphas,
            "betas": None if self.betas is None else self.betas,
        }
        return {key: val for key, val in groups.items() if val is not None}

    @cached_property
    def as_list(self) -> list[torch.Tensor]:
        """Gets a list of all the unique parameters.

        Returns:
            list[torch.Tensor]: The list of the (unique) parameters.
        """
        seen: set[torch.Tensor] = set()
        _is_new = lambda x: not (x in seen or seen.add(x))  # noqa: E731  # type: ignore

        def _items(
            v: torch.Tensor | dict[tuple[Any, Any], torch.Tensor],
        ) -> list[torch.Tensor]:
            if isinstance(v, torch.Tensor):
                return [v] if _is_new(v) else []
            return [t for t in v.values() if _is_new(t)]

        return list(chain.from_iterable(_items(v) for v in self.as_dict.values()))

    @cached_property
    def as_named_list(self) -> list[tuple[str, torch.Tensor]]:
        """Gets a list of all the unique parameters names and values.

        Returns:
            list[tuple[str, torch.Tensor]]: The list of the (unique) parameters names
                and values.
        """
        seen: set[torch.Tensor] = set()
        _is_new = lambda x: not (x in seen or seen.add(x))  # noqa: E731  # type: ignore

        def _items(
            k: str, v: torch.Tensor | dict[tuple[Any, Any], torch.Tensor]
        ) -> list[tuple[str, torch.Tensor]]:
            if isinstance(v, torch.Tensor):
                return [(k, v)] if _is_new(v) else []
            return [(f"{k}[{sk}]", sv) for sk, sv in v.items() if _is_new(sv)]

        return list(chain.from_iterable(_items(k, v) for k, v in self.as_dict.items()))

    @property
    def as_flat_tensor(self) -> Tensor1D:
        """Get the flattened unique parameters.

        Returns:
            torch.Tensor: The flattened (unique) parameters.
        """
        return torch.cat([p.reshape(-1) for p in self.as_list])

    @cached_property
    def numel(self) -> int:
        """Return the number of unique parameters.

        Returns:
            int: The number of the (unique) parameters.
        """
        return sum(p.numel() for p in self.as_list)

    def requires_grad_(self, req: bool):
        """Enable or disable gradient computation on non extra parameters.

        Args:
            req (bool): Wether to require or not.
        """
        for t in self.as_list:
            t.requires_grad_(req)

    def extra_requires_grad_(self, req: bool):
        """Enable or disable gradient computation on extra parameters.

        Args:
            req (bool): Wether to require or not.
        """
        if self.extra is None:
            return
        for t in self.extra:
            t.requires_grad_(req)

    def get_cov(self, matrix: str) -> torch.Tensor:
        """Get covariance from parameter.

        Args:
            matrix (str): Either "Q" or "R".

        Raises:
            ValueError: If the matrix is not in ("Q", "R")

        Returns:
            torch.Tensor: The covariance matrix.
        """
        if matrix not in ("Q", "R"):
            raise ValueError(f"matrix must be either Q or R, got {matrix}")

        # Get repr then covariance
        return cov_from_repr(getattr(self, matrix + "_repr"))

    @validate_call(config=ConfigDict(arbitrary_types_allowed=True))
    def from_flat_tensor(self, flat: Tensor1D) -> Self:
        """Gets a ModelParams object based on the flat representation.

        This uses the current object as the reference.

        Args:
            flat (torch.Tensor): The flat representation.

        Returns:
            Self The constructed ModelParams.
        """
        seen: dict[torch.Tensor, torch.Tensor] = {}
        i = 0

        def _next(ref: torch.Tensor):
            nonlocal seen, i
            if ref in seen:
                return seen[ref]

            n = ref.numel()
            result = flat[i : i + n]
            i += n
            seen[ref] = result.reshape(ref.shape)

            return seen[ref]

        return cast(Self, map_fn_params(self, _next))

    def detach(self) -> Self:
        """Returns a detached reshape of the parameters.

        Returns:
            Self The detached reshape.
        """
        return cast(Self, map_fn_params(self, torch.detach))

    def clone(self) -> Self:
        """Returns a clone of the parameters.

        Returns:
            Self The clone.
        """
        seen: dict[torch.Tensor, torch.Tensor] = {}

        def _next(t: torch.Tensor):
            nonlocal seen
            if t not in seen:
                seen[t] = t.clone()

            return seen[t]

        return cast(Self, map_fn_params(self, _next))



import warnings
from collections.abc import Callable
from typing import Any

import torch
from pydantic import ConfigDict, validate_call

from ..typedefs._defs import Info, Job, Metrics
from ..utils._dtype import get_dtype


class ComputeFIM(Job):
    r"""Job to compute the Fisher Information Matrix.

    Many methods exist for computing the Fisher Information Matrix in latent variable
    models. In particular, this class leverages the expected Fisher Information Matrix
    using the identity:

    .. math::
        \mathcal{I}(\theta) = \mathbb{E}_{b \sim p(\cdot \mid x, \theta)} \left(\nabla
        \log \mathcal{L}(x, b ; \theta) \nabla \log \mathcal{L}(x, b ; \theta)^T \right)

    Please note that this is a stochastic approximation. By using the `bias=True`
    option which is enabled by default, you can substract a bias term giving the
    covariance matrix instead. This can be useful if the parameter estimate is quite
    far from the MLE.

    For more information, see PMLR 202:1430-1453, 2023.

    Attributes:
        grad_m1 (torch.Tensor): The moment of order 1 estimate of the gradient.
        grad_m2 (torch.Tensor): The moment of order 2 estimate of the gradient.
        jac_fn (Callable[[torch.Tensor, torch.Tensor], torch.Tensor]): The Jacobian
            function. Useful because of batching.
    """

    grad_m1: torch.Tensor
    grad_m2: torch.Tensor
    jac_fn: Callable[[torch.Tensor, torch.Tensor], torch.Tensor]

    @validate_call(config=ConfigDict(arbitrary_types_allowed=True))
    def __new__(cls, bias: bool = True) -> Callable[[Info], Job]:
        """Creates the Fisher information Job.

        Args:
            bias (bool): Whether or not to substract the bias term. Defaults to True.
        """
        return super().__new__(cls, bias)

    def __init__(self, bias: bool, *, info: Info):  # type: ignore
        """Initializes the moment estimates to 0.

        Warns the user if the model has not yet been fitted through the `fit`
        attribute.

        Args:
            bias (bool): Whether or not to substract the bias term. Defaults to True.
            info (Info): The job information object.
        """
        if not info.model.fit_:
            warnings.warn(
                "Model should be fitted before computing Fisher Information Matrix",
                stacklevel=2,
            )

        dtype = get_dtype()

        d = info.model.params_.numel
        self.grad_m2 = torch.zeros(d, d, dtype=dtype)
        if bias:
            self.grad_m1 = torch.zeros(d, dtype=dtype)

        def _jac_fn(params_flat_tensor: torch.Tensor, b: torch.Tensor):
            params = info.model.params_.from_flat_tensor(params_flat_tensor)
            return info.logpdfs_aux_fn(params, b)[0].sum(dim=1)

        self.jac_fn = torch.func.jacrev(_jac_fn)  # type: ignore

    def run(self, info: Info):
        """Updates the moments estimates by stochastic approximation.

        Args:
            info (Info): The job information object.
        """
        jac = self.jac_fn(info.model.params_.as_flat_tensor, info.sampler.b).detach()

        self.grad_m2.addmm_(jac.T, jac, alpha=1.0 / info.sampler.n_chains)
        if hasattr(self, "grad_m1"):
            self.grad_m1 += jac.mean(dim=0)

    def end(self, info: Info, metrics: Metrics):
        """Writes the Fisher Information Matrix into metrics.

        Args:
            info (Info): The job information object.
            metrics (Metrics): The job metrics object.
        """
        self.grad_m2 /= info.iteration
        if hasattr(self, "grad_m1"):
            self.grad_m1 /= info.iteration

        metrics.fim = (
            self.grad_m2 - torch.outer(self.grad_m1, self.grad_m1)
            if hasattr(self, "grad_m1")
            else self.grad_m2
        )


class ComputeCriteria(Job):
    r"""Job to compute AIC, BIC and log likelihood.

    These criteria allow for model selection.

    The AIC criterion is given by:

    .. math::
        \text{AIC} = -2 \log \mathcal{L}(x) + 2 k,

    where :math:`k` is the number of parameters.

    The BIC criterion is given by:

    .. math::
        \text{BIC} = -2 \log \mathcal{L}(x) + 2 \log n,

    where :math:`n` is the number of effective data measurements. Note this is computed
    as the sum of the lengths of the trajectories added to the sum of longitudinal data
    measurements. If the residuals are assumed to be independent, then the longitudinal
    contribution is equal to the number of all non NaN values given in `data.y`.
    Otherwise, it is equal to the number of measurement vectors provided.

    Please note that this is a stochastic approximation.

    Attributes:
        loglik (float): The log likelihood.
    """

    loglik: float

    def __new__(cls) -> Callable[[Info], Job]:
        """Creates the criteria computation job."""
        return super().__new__(cls)

    def __init__(self, info: Info, **_kwargs: Any):  # type: ignore
        """Initializes the log likelihood to 0.

        Args:
            info (Info): The job information object.
        """
        if not info.model.fit_:
            warnings.warn(
                "Model should be fitted before computing criteria", stacklevel=2
            )

        self.loglik = 0.0

    def run(self, info: Info):
        """Updates the log likelihood by stochastic approximation.

        Args:
            info (Info): The job information object.
        """
        self.loglik += info.sampler.aux.logliks.sum().item() / info.sampler.n_chains

    def end(self, info: Info, metrics: Metrics):
        """Computes other criteria and write them into metrics.

        Args:
            info (Info): The job information object.
            metrics (Metrics): The job metrics object.
        """
        metrics.loglik = self.loglik / info.iteration
        metrics.nloglik_pen = (
            info.data.size * info.model.pen(info.model.params_).item() - metrics.loglik
            if info.model.pen is not None
            else -metrics.loglik
        )

        # Compute penalties; if the residuals are assumed to be independent, the
        # effective sample size changes
        d = info.model.params_.numel
        n = info.data.effective_size(info.model.params_.R_repr.method != "full")
        aic_pen = 2 * d
        bic_pen = d * torch.log(torch.tensor(n)).item()

        metrics.aic = 2 * metrics.nloglik_pen + aic_pen
        metrics.bic = 2 * metrics.nloglik_pen + bic_pen


class ComputeEBEs(Job):
    r"""Job to compute the EBEs of b.

    Empirical Bayes Estimators are given by the posterior means of the random effects:

    .. math::
        \text{EBE} = \mathbb{E}_{b \sim p(\cdot \mid x, \theta)}(b).

    Please note that this is a stochastic approximation.

    Attributes:
        ebes (torch.Tensor): The Empirical Bayes Estimators.
    """

    ebes: torch.Tensor

    def __new__(cls) -> Callable[[Info], Job]:
        """Creates the EBEs computation job."""
        return super().__new__(cls)

    def __init__(self, info: Info):  # type: ignore
        """Initializes the EBEs to 0.

        Args:
            info: The job information object.
        """
        self.ebes = torch.zeros(info.sampler.b.shape[1:], dtype=get_dtype())

    def run(self, info: Info):
        """Updates the EBEs by stochastic approximation.

        Args:
            info (Info): The job information object.
        """
        self.ebes += info.sampler.b.mean(dim=0)

    def end(self, info: Info, metrics: Metrics):
        """Writes EBEs into metrics.

        Args:
            info (Info): The job information object.
            metrics (Metrics): The job metrics object.
        """
        metrics.ebes = self.ebes / info.iteration



from abc import ABC, abstractmethod
from collections.abc import Callable
from typing import Any, Final

import torch
from pydantic import ConfigDict, validate_call

from ..typedefs._defs import Info, Job, NumNonNegative

ADAM_LIKE: Final[tuple[type[torch.optim.Optimizer], ...]] = (
    torch.optim.Adam,
    torch.optim.AdamW,
    torch.optim.NAdam,
)


class _BaseL1Proximal(Job, ABC):
    """Base class for proximal operators.

    Attributes:
        lmda (int | float): The penalty.
        group (str): The group to penalize, either the link or covariate
            parameters.
        unique_params (list[torch.Tensor]): The unique parameters to penalize.
    """

    group: str
    lmda: int | float
    unique_params: list[torch.Tensor]

    @validate_call(config=ConfigDict(arbitrary_types_allowed=True))
    def __new__(
        cls, lmda: NumNonNegative, group: str = "betas"
    ) -> Callable[[Info], Job]:
        """Creates the proximal operator.

        Args:
            lmda (NumNonNegative): The penalty.
            group (str, optional): The group to penalize, either the link or covariate
                parameters. Defaults to "betas".
        """
        return super().__new__(cls, lmda, group)

    def __init__(self, lmda: NumNonNegative, group: str = "betas", *, info: Info):  # type: ignore
        """Initialize the proximal operator.

        Args:
            lmda (NumNonNegative): The penalty.
            group (str, optional): The group to penalize, either the link or covariate
                parameters. Defaults to "betas".
            info (Info): The job information object.

        Raises:
            ValueError: If the optimizer has not been initialized before the proximal.
            ValueError: If the group is not in `alphas` nor `betas`.
            ValueError: If the group is None.
        """
        self.group = group
        self.lmda = lmda
        if group not in ("alphas", "betas"):
            raise ValueError(
                f"Group must be either 'alphas' or 'betas', got {self.group}"
            )

        if not hasattr(info, "opt"):
            raise ValueError("Optimizer must be initialized before proximal job")
        if getattr(info.model.params_, self.group) is None:
            raise ValueError(f"{self.group} is None")

        self.check_optimizer(info.opt)

        self.unique_params = list(set(getattr(info.model.params_, group).values()))

    def run(self, info: Info):
        """Projects the current parameter value.

        Args:
            info (Info): The job information object.
        """
        g = info.opt.param_groups[0]
        for p in self.unique_params:
            if p.grad is None:
                continue

            state = info.opt.state[p]
            if len(state) == 0:
                continue

            eff_lr = self.get_effective_lr(g, state)
            p.data = p.sign() * torch.clamp(p.abs() - self.lmda * eff_lr, min=0.0)

    @staticmethod
    @abstractmethod
    def check_optimizer(optimizer: torch.optim.Optimizer):
        """Checks if the optimizer is supported.

        Args:
            optimizer (torch.optim.Optimizer): The optimizer object.
        """

    @staticmethod
    @abstractmethod
    def get_effective_lr(g: dict[str, Any], state: dict[str, Any]) -> float:
        """Get the effective learning rate.

        Args:
            g (dict[str, Any]): The parameter group/
            state (dict[str, Any]): The optimizer state.
        """


class AdamL1Proximal(_BaseL1Proximal):
    r"""Adam proximal operator.

    This proximal operator aims at bringing variable selection to joint modeling.
    This proximal operator only works with Adam or Adam-like optimizers that compute
    exponential moving averages of the order 1 and 2 moments of the gradient.

    Mathematically, consider the debiased estimates of the gradient and its element-wise
    square:

    .. math::
        \hat{m}_1^{(t)}, \quad \hat{m}_2^{(t)}.

    The proximal operator is given by the following formula. If :math:`\theta` are the
    model parameters, define for a L1 penalty :math:`\lambda \geq 0`:

    .. math::
        \theta_\text{pre}^{(t+1)} \gets \theta^{(t+1)} + \frac{\text{lr}}{\epsilon +
        \sqrt{\hat{m}_2^{(t)}}} \hat{m}_1^{(t)},

    then compute the projection:

    .. math::
        \theta^{(t+1)} \gets \operatorname{Proxi}_{\lambda \frac{\text{lr}}{\epsilon
        + \sqrt{\hat{m}_2^{(t)}}}}(\theta_\text{pre}^{(t+1)}).

    The proximal operator is defined by the element-wise soft thresholding:

    .. math::
        \theta_i^{(t+1)} \gets \operatorname{sgn}(\theta_{\text{pre}, i}^{(t+1)})
        \max\bigl(\vert \theta_{\text{pre}, i}^{(t+1)} \vert - \lambda \frac{\text{lr}}
        {\epsilon + \sqrt{\hat{m}_{2, i}^{(t)}}}, 0\bigr).

    Attributes:
        lmda (int | float): The penalty.
        group (str): The group to penalize, either the link or covariate
            parameters.
        unique_params (list[torch.Tensor]): The unique parameters to penalize.
    """

    @staticmethod
    def check_optimizer(optimizer: torch.optim.Optimizer):
        """Checks if the optimizer is supported.

        Args:
            optimizer (torch.optim.Optimizer): The optimizer object.

        Raises:
            ValueError: If the optimizer is not Adam-like.
        """
        if not isinstance(optimizer, ADAM_LIKE):
            raise ValueError("Optimizer must be Adam or Adam-like")

    @staticmethod
    def get_effective_lr(g: dict[str, Any], state: dict[str, Any]) -> float:
        """Gets the effective learning rate.

        Args:
            g (dict[str, Any]): The parameter group.
            state (dict[str, Any]): The optimizer state.

        Returns:
            float: The effective learning rate.
        """
        return g["lr"] / torch.sqrt(state["exp_avg_sq"] + g["eps"])



from dataclasses import replace
from typing import Any, Final

import torch
from xxhash import xxh3_64_intdigest

from ..typedefs._data import CompleteModelData, ModelDesign, SampleData
from ..typedefs._defs import HazardInfo, Trajectory
from ..typedefs._params import ModelParams
from ..utils._dtype import get_dtype
from ..utils._misc import legendre_quad
from ..utils._surv import build_possible_buckets, build_remaining_buckets
from ._cache import Cache

# Constants
HAZARD_CACHE_KEYS: Final[tuple[str, ...]] = ("base", "half", "quad_c", "quad_lc")


class HazardMixin:
    """Mixin class for hazard model computations."""

    params_: ModelParams
    model_design: ModelDesign
    n_quad: int
    n_bisect: int
    cache_limit: int | None
    _std_nodes: torch.Tensor
    _std_weights: torch.Tensor
    _cache: Cache

    def __init__(
        self,
        n_quad: int,
        n_bisect: int,
        cache_limit: int | None,
        *args: Any,
        **kwargs: Any,
    ):
        """Initializes the hazard mixin.

        Args:
            n_quad (int): Number of quadrature nodes.
            n_bisect (int): The number of bisection steps.
            cache_limit (int | None): Max length of cache.
        """
        self.n_quad = n_quad
        self.n_bisect = n_bisect
        self.cache_limit = cache_limit
        self._std_nodes, self._std_weights = legendre_quad(n_quad)
        self._cache = Cache(cache_limit, HAZARD_CACHE_KEYS)

        super().__init__(*args, **kwargs)

    def _log_hazard(self, hazard_info: HazardInfo) -> torch.Tensor:
        """Computes log hazard.

        Args:
            hazard_info (HazardInfo): All necessary information for computation.

        Returns:
            torch.Tensor: The computed log hazard.
        """
        # Unpack data
        t0, t1, x, psi, alpha, beta, base_hazard_fn, link_fn = hazard_info

        # Compute baseline hazard
        def _base_create():
            return base_hazard_fn(t0, t1)

        if self.cache_limit != 0:
            try:
                key = (
                    *hazard_info.base_hazard_fn.key,
                    xxh3_64_intdigest(t0.detach().contiguous().numpy()),  # type: ignore
                    xxh3_64_intdigest(t1.detach().contiguous().numpy()),  # type: ignore
                )
                base = self._cache.get_cache("base", key, _base_create)
            except RuntimeError:
                base = _base_create()
        else:
            base = _base_create()

        # Compute time-varying effects
        mod = link_fn(t1, psi) @ alpha

        # Compute covariates effect
        if x is None or beta is None:
            return base + mod

        return base + mod + x @ beta.unsqueeze(-1)

    def _cum_hazard(
        self, hazard_info: HazardInfo, c: torch.Tensor | None
    ) -> torch.Tensor:
        """Computes cumulative hazard.

        Args:
            hazard_info (HazardInfo): All necessary information for computation.
            c (torch.Tensor | None): Integration start or censoring time, t0 if None.

        Returns:
            torch.Tensor: The computed cumulative hazard.
        """
        # Unpack data
        t0, t1, *_ = hazard_info
        c = t0 if c is None else c

        # Transform to quadrature interval
        def _half_create():
            return 0.5 * (t1 - c)

        def _quad_c_create():
            return (
                (c + t1)
                .reshape(-1, 1)
                .addmm(half.reshape(-1, 1), self._std_nodes, beta=0.5)
                .reshape(t1.size(0), -1)
            )

        if self.cache_limit != 0:
            try:
                key = (
                    xxh3_64_intdigest(c.detach().contiguous().numpy()),  # type: ignore
                    xxh3_64_intdigest(t1.detach().contiguous().numpy()),  # type: ignore
                )
                half = self._cache.get_cache("half", key, _half_create)
                quad_c = self._cache.get_cache("quad_c", key, _quad_c_create)
            except RuntimeError:
                half = _half_create()
                quad_c = _quad_c_create()
        else:
            half = _half_create()
            quad_c = _quad_c_create()

        # Compute hazard at quadrature points
        vals = self._log_hazard(hazard_info._replace(t1=quad_c))
        vals.clamp_(max=50.0).exp_()

        return half.reshape(t1.shape) * (
            vals.reshape(*vals.shape[:-1], -1, self._std_weights.size(-1))
            @ self._std_weights
        )

    def _log_and_cum_hazard(
        self,
        hazard_info: HazardInfo,
        enable_cache: bool,
    ) -> tuple[torch.Tensor, torch.Tensor]:
        """Computes both log and cumulative hazard.

        Args:
            hazard_info (HazardInfo): All necessary information for computation.
            enable_cache (bool): Enables caching.

        Raises:
            RuntimeError: If the computation fails.

        Returns:
            tuple[torch.Tensor, torch.Tensor]: The log and cum hazard.
        """
        # Unpack data
        t0, t1, *_ = hazard_info

        # Transform to quadrature interval
        def _half_create():
            return 0.5 * (t1 - t0)

        def _quad_lc_create():
            return torch.cat(
                [t1, (t0 + t1).addmm(half, self._std_nodes, beta=0.5)], dim=-1
            )

        if enable_cache:
            try:
                key = (
                    xxh3_64_intdigest(t0.detach().contiguous().numpy()),  # type: ignore
                    xxh3_64_intdigest(t1.detach().contiguous().numpy()),  # type: ignore
                )
                half = self._cache.get_cache("half", key, _half_create)
                quad_lc = self._cache.get_cache("quad_c", key, _quad_lc_create)
            except RuntimeError:
                half = _half_create()
                quad_lc = _quad_lc_create()
        else:
            half = _half_create()
            quad_lc = _quad_lc_create()

        # Compute log hazard at all points
        all_vals = self._log_hazard(hazard_info._replace(t1=quad_lc))
        all_vals[..., 1:].clamp_(max=50.0).exp_()

        return all_vals[..., 0], half.reshape(-1) * (
            all_vals[..., 1:] @ self._std_weights
        )

    def _sample_transition(
        self,
        hazard_info: HazardInfo,
        c: torch.Tensor | None,
    ) -> torch.Tensor:
        """Sample survival times using inverse transform sampling.

        Args:
            hazard_info (HazardInfo): All necessary information for computation.
            c (torch.Tensor | None): Conditionning time.

        Returns:
            torch.Tensor: The computed pre transition times.
        """
        # Unpack data
        t0, t1, *_ = hazard_info

        # Initialize for bisection search
        t_left, t_right = (
            t0.clone(),
            torch.nextafter(t1, torch.tensor(torch.inf, dtype=t1.dtype)),
        )

        # Generate exponential random variables
        target = -torch.log(torch.rand_like(t_left))

        # Bisection search for survival times
        for _ in range(self.n_bisect):
            t_mid = 0.5 * (t_left + t_right)

            surv_logps = self._cum_hazard(hazard_info._replace(t1=t_mid), c)

            # Update search bounds
            accept_mask = surv_logps.reshape(target.shape) < target
            torch.where(accept_mask, t_mid, t_left, out=t_left)
            torch.where(accept_mask, t_right, t_mid, out=t_right)

        return t_right

    def _sample_trajectory_step(
        self, sample_data: SampleData, c_max: torch.Tensor
    ) -> bool:
        """Appends the next simulated transition.

        Args:
            sample_data (SampleData): Sampling data
            c_max (torch.Tensor): Max sampling time.

        Returns:
            bool: If the sampling is done.
        """
        # Unpack data
        x = sample_data.x
        psi = sample_data.psi
        c = sample_data.c

        # Get buckets from last states
        current_buckets = build_possible_buckets(
            sample_data.trajectories, c_max, tuple(self.model_design.surv.keys())
        )

        if not current_buckets:
            return False

        # Initialize candidate transition times
        n_transitions = len(current_buckets)
        t_candidates = torch.full(
            (sample_data.size, n_transitions), torch.inf, dtype=get_dtype()
        )

        for j, (key, bucket) in enumerate(current_buckets.items()):
            idxs, t0, t1 = bucket

            # Create info
            hazard_info = HazardInfo(
                t0,
                t1,
                None if x is None else x.index_select(-2, idxs),
                psi.index_select(-2, idxs),
                self.params_.alphas[key],
                None if self.params_.betas is None else self.params_.betas[key],
                *self.model_design.surv[key],
            )

            # Sample transition times
            t_sample = self._sample_transition(
                hazard_info,
                None if c is None else c.index_select(-2, idxs),
            )

            t_candidates[idxs, j] = t_sample.reshape(-1)

        # Find earliest transition
        min_times, argmin_idxs = torch.min(t_candidates, dim=1)
        bucket_keys = list(current_buckets.keys())

        for i, (time, arg_idx) in enumerate(zip(min_times, argmin_idxs, strict=True)):
            if torch.isfinite(time):
                sample_data.trajectories[i].append(
                    (time.item(), bucket_keys[int(arg_idx)][1])
                )

        return True

    def _sample_trajectories(
        self,
        sample_data: SampleData,
        c_max: torch.Tensor,
        max_length: int,
    ) -> list[Trajectory]:
        """Sample future trajectories from the fitted joint model.

        Args:
            sample_data (SampleData): Prediction data.
            c_max (torch.Tensor): The maximum trajectory censoring times.
            max_length (int): Maximum iterations or sampling.

        Returns:
            list[Trajectory]: The sampled trajectories.
        """
        sample_data_copied = replace(sample_data, skip_validation=True)

        # Sample future transitions iteratively
        for _ in range(max_length):
            if not self._sample_trajectory_step(
                sample_data_copied,
                c_max,
            ):
                break
            object.__setattr__(sample_data_copied, "c", None)

        return [
            trajectory if trajectory[-1][0] <= c_max[i] else trajectory[:-1]
            for i, trajectory in enumerate(sample_data_copied.trajectories)
        ]

    def _compute_surv_logps(
        self, sample_data: SampleData, u: torch.Tensor
    ) -> torch.Tensor:
        """Computes log probabilites of remaining event free up to time u.

        Args:
            sample_data (SampleData): The data on which to compute the probabilities.
            u (torch.Tensor): The time at which to evaluate the probabilities.

        Returns:
            torch.Tensor: The computed survival log probabilities.
        """
        # Unpack data
        x = sample_data.x
        psi = sample_data.psi
        c = sample_data.c

        # Get buckets from last states
        dtype = get_dtype()
        buckets = build_remaining_buckets(
            sample_data.trajectories, tuple(self.model_design.surv.keys())
        )

        nlogps = torch.zeros(*psi.shape[:-1], u.size(1), dtype=dtype)

        # Compute the log probabilities summing over transitions
        for key, bucket in buckets.items():
            idxs, t0 = bucket

            # Create info
            hazard_info = HazardInfo(
                t0,
                u.index_select(0, idxs),
                None if x is None else x.index_select(-2, idxs),
                psi.index_select(-2, idxs),
                self.params_.alphas[key],
                None if self.params_.betas is None else self.params_.betas[key],
                *self.model_design.surv[key],
            )

            # Compute negative log survival
            alts_logliks = self._cum_hazard(
                hazard_info, None if c is None else c.index_select(0, idxs)
            )

            nlogps.index_add_(-2, idxs, alts_logliks)

        return -nlogps.clamp(min=0.0)

    def _hazard_logliks(
        self, params: ModelParams, psi: torch.Tensor, data: CompleteModelData
    ) -> torch.Tensor:
        """Computes the hazard log likelihoods.

        Args:
            params (ModelParams): The model parameters.
            psi (torch.Tensor): A matrix of individual parameters.
            data (CompleteModelData): Dataset on which likelihood is computed.
            enable_cache (bool): Enable caching

        Returns:
            torch.Tensor: The computed log likelihoods.
        """
        logliks = torch.zeros(psi.shape[:-1], dtype=get_dtype())

        for key, bucket in data.buckets.items():
            idxs, t0, t1, obs = bucket

            # Create info
            hazard_info = HazardInfo(
                t0,
                t1,
                None if data.x is None else data.x.index_select(-2, idxs),
                psi.index_select(-2, idxs),
                params.alphas[key],
                None if params.betas is None else params.betas[key],
                *self.model_design.surv[key],
            )

            obs_logliks, alts_logliks = self._log_and_cum_hazard(
                hazard_info,
                self.cache_limit != 0,
            )

            vals = (-alts_logliks).addcmul(obs, obs_logliks)
            logliks.index_add_(-1, idxs, vals)

        return logliks



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "jmstate"
release = ""
version = ""
copyright = "2025, Félix Laplante"
author = "Félix Laplante"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "sphinx.ext.autosummary",
]

templates_path = ["_templates"]
exclude_patterns = []

autodoc_member_order = "bysource"
autodoc_typehints = "description"
autodoc_typehints_format = "short"
autodoc_inherit_docstrings = True
autosummary_generate = True
add_module_names = False
napoleon_use_ivar = True
napoleon_attr_annotations = True

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "furo"
html_static_path = ["_static"]



from typing import Any

import torch

from ..typedefs._data import CompleteModelData, ModelDesign
from ..typedefs._defs import LOG_TWO_PI
from ..typedefs._params import ModelParams
from ..utils._linalg import get_cholesky_and_log_eigvals


class LongitudinalMixin:
    """Mixin class for longitudinal model computations."""

    model_design: ModelDesign

    def __init__(self, *args: Any, **kwargs: Any):
        """Initializes the longitudinal mixin."""
        super().__init__(*args, **kwargs)

    def _long_logliks(
        self, params: ModelParams, psi: torch.Tensor, data: CompleteModelData
    ) -> torch.Tensor:
        """Computes the longitudinal log likelihoods.

        Args:
            params (ModelParams): The model parameters.
            psi (torch.Tensor): A 3D tensor of individual parameters.
            data (ModelData): Dataset on which likelihood is computed.

        Returns:
            torch.Tensor: The computed log likelihoods.
        """
        # Careful with NaNs
        predicted = self.model_design.regression_fn(data.valid_t, psi)
        diffs = data.valid_y.addcmul(predicted, data.valid_mask, value=-1.0)

        R_inv_cholesky, R_nlog_eigvals = get_cholesky_and_log_eigvals(params, "R")
        R_quad_forms = (diffs @ R_inv_cholesky).pow(2).sum(dim=(-2, -1))
        R_norm_factor = data.n_valid @ (R_nlog_eigvals - LOG_TWO_PI)

        return 0.5 * (R_norm_factor - R_quad_forms)



import warnings
from abc import ABC, abstractmethod
from collections.abc import Callable
from functools import wraps
from typing import Any

import torch
from pydantic import ConfigDict, validate_call

from ..typedefs._defs import (
    Info,
    IntStrictlyPositive,
    Job,
    NumNonNegative,
    NumProbability,
    Tensor1DPositive,
)
from ..utils._checks import check_consistent_size
from ..utils._dtype import get_dtype

# Constants
NOT_CONVERGED_WARNING = (
    "The parameters are not converged, try increasing the maximum number "
    "of iterations, decreasing the learning rate, or increasing n_chains"
)


class NoStop(Job):
    """Job to disallow convergence until a minimum number of iterations is reached.

    All stopping jobs have to agree and return True for the process to stop. Other
    jobs return None.

    Attributes:
        min_steps (int): The minimum number of steps to allow convergence.
    """

    min_steps: int

    @validate_call(config=ConfigDict(arbitrary_types_allowed=True))
    def __new__(cls, min_steps: IntStrictlyPositive) -> Callable[[Info], Job]:
        """Creates the `NoStop` class.

        Args:
            min_steps (IntStrictlyPositive): The minimum number of steps before allowing
                convergence checking.
        """
        return super().__new__(cls, min_steps)

    def __init__(  # type: ignore
        self, min_steps: IntStrictlyPositive, **_kwargs: Any
    ):
        """Initializes the `NoStop` class.

        Args:
            min_steps (IntStrictlyPositive): The minimum number of steps before allowing
                convergence checking.
        """
        self.min_steps = min_steps

    def run(self, info: Info) -> bool | None:
        """Returns True if convergence checking is allowed, else None.

        Args:
            info (Info): The job information object.

        Returns:
            bool: True if convergence checking is allowed, else None.
        """
        return None if info.iteration >= self.min_steps else False


class _BaseEMAStop(Job, ABC):
    """Job base class to test the convergence based on exponential moving averages.

    Attributes:
        atol (NumNonNegative | Tensor1DPositive, optional): Absolute tolerance,
            either scalar or element-wise.
        rtol (NumNonNegative | Tensor1DPositive, optional): Relative tolerance,
        either scalar or element-wise.
        min_consecutive (IntStrictlyPositive, optional): The minimum consecutive
            iterations with grad difference less than tolerance.
        betas (tuple[NumProbability, NumProbability], optional): Exponential moving
            averages' forget parameters.
        m (torch.Tensor): The first moment estimate.
        v (torch.Tensor): The second moment estimate.
        n_consecutive (int): The number of iterations with convergence criterion met.
    """

    atol: int | float | torch.Tensor
    rtol: int | float | torch.Tensor
    min_consecutive: int
    betas: tuple[int | float, int | float]
    m: torch.Tensor
    v: torch.Tensor
    n_consecutive: int

    @validate_call(config=ConfigDict(arbitrary_types_allowed=True))
    def __new__(
        cls,
        atol: NumNonNegative | Tensor1DPositive = 1e-6,
        rtol: NumNonNegative | Tensor1DPositive = 0.1,
        min_consecutive: IntStrictlyPositive = 1,
        *,
        betas: tuple[NumProbability, NumProbability] = (0.99, 0.99),
    ) -> Callable[[Info], Job]:
        """Creates the class.

        Args:
            atol (NumNonNegative | Tensor1DPositive, optional): Absolute tolerance,
            either scalar or element-wise. Defaults to 1e-6.
            rtol (NumNonNegative | Tensor1DPositive, optional): Relative tolerance,
            either scalar or element-wise. Defaults to 0.1.
            min_consecutive (IntStrictlyPositive, optional): The minimum consecutive
                iterations with grad difference less than tolerance. Defaults to 1.
            betas (tuple[NumProbability, NumProbability], optional): Exponential moving
                averages' forget parameters. Defaults to (0.99, 0.99).
        """
        return super().__new__(cls, atol, rtol, min_consecutive, betas=betas)

    def __init__(  # type: ignore
        self,
        atol: NumNonNegative | Tensor1DPositive = 1e-6,
        rtol: NumNonNegative | Tensor1DPositive = 0.1,
        min_consecutive: IntStrictlyPositive = 1,
        *,
        betas: tuple[NumProbability, NumProbability] = (0.99, 0.99),
        **_kwargs: Any,
    ):
        """Initializes the class.

        Args:
            atol (NumNonNegative | Tensor1DPositive, optional): Absolute tolerance,
            either scalar or element-wise. Defaults to 1e-6.
            rtol (NumNonNegative | Tensor1DPositive, optional): Relative tolerance,
            either scalar or element-wise. Defaults to 0.1.
            min_consecutive (IntStrictlyPositive, optional): The minimum consecutive
                iterations with grad difference less than tolerance. Defaults to 1.
            betas (tuple[NumProbability, NumProbability], optional): Exponential moving
                averages' forget parameters. Defaults to (0.99, 0.99).

        Raises:
            ValueError: If the shapes of the tolerances are not compatible.
        """
        self.atol = atol
        self.rtol = rtol
        self.min_consecutive = min_consecutive
        self.betas = betas

        self.n_consecutive = 0

    def run(self, info: Info) -> bool:
        """Updates the moment estimates stochastically and checks for convergence.

        Args:
            info (Info): The job information object.

        Returns:
            bool: True if parameters have convergence, else False.
        """
        grads_list: list[torch.Tensor] = []
        for p in info.model.params_.as_list:
            if p.grad is None:
                grads_list.append(torch.zeros_like(p))
            else:
                grads_list.append(p.grad)

        q = self.quantity(info)
        self.m = self.betas[0] * self.m + (1 - self.betas[0]) * q
        self.v = self.betas[0] * self.v + (1 - self.betas[1]) * q**2

        m_hat = self.m / (1 - self.betas[0] ** (info.iteration + 1))
        v_hat = self.v / (1 - self.betas[1] ** (info.iteration + 1))

        # Check convergence
        if (m_hat.abs() <= self.atol + self.rtol * v_hat.sqrt()).all():
            self.n_consecutive += 1
        else:
            self.n_consecutive = 0

        return self.n_consecutive >= self.min_consecutive

    def end(self, **_kwargs: Any):
        """Checks and warns the user if the parameters have not yet converged.

        All stopping jobs have to agree and return True for the process to stop. Other
        jobs return None.
        """
        if self.n_consecutive < self.min_consecutive:
            warnings.warn(
                NOT_CONVERGED_WARNING,
                stacklevel=2,
            )

    @abstractmethod
    def quantity(self, info: Info) -> torch.Tensor:
        """Gets the interest quantity.

        Returns:
            torch.Tensor: The quantity of interest.
        """


class GradStop(_BaseEMAStop):
    r"""Job to test the convergence based on gradient.

    All stopping jobs have to agree and return True for the process to stop. Other
    jobs return None.

    Mathematically, estimates of the first and second moments of the gradient are
    stochastically computed, using exponential moving averages with the formula:

    .. math::
        m_i^{(t)} \gets \beta_i m_i^{(t-1)} + (1 - \beta_i) \bigl(\nabla \log
        \mathcal{L}(\theta^{(t)})\bigr)^i, \quad \hat{m}_i^{(t)} = \frac{m_i^{(t)}}
        {1 - \beta_i^t}.

    The `run` method returns true when for `min_consecutive` iterations:

    .. math::
        \vert \hat{m}_1^{(t)} \vert \leq \text{atol} + \text{rtol} \odot
        \sqrt{\hat{m}_2^{(t)}}.

    Please note the tolerances must both be non-negative and can represent element-wise
    tolerances or global tolerance common to all parameters.

    Attributes:
        atol (int | float | torch.Tensor): Absolute tolerance.
        rtol (int | float | torch.Tensor): Relative tolerance.
        min_consecutive (int): Minimum consecutive iterations to declare convergence.
        betas (tuple[int | float, int | float]): The forget parameters
        m (torch.Tensor): The first moment estimate.
        v (torch.Tensor): The second moment estimate.
        n_consecutive (int): The number of iterations with convergence criterion met.
        stopped (bool): Indicator whether or not convergence has happened.
    """

    @wraps(_BaseEMAStop.__init__)
    def __init__(self, *args: Any, info: Info, **kwargs: Any):
        super().__init__(*args, **kwargs)

        d = info.model.params_.numel
        if isinstance(self.atol, torch.Tensor):
            check_consistent_size(((self.atol, 0, "atol"), (d, None, "params.numel")))
        if isinstance(self.rtol, torch.Tensor):
            check_consistent_size(((self.rtol, 0, "rtol"), (d, None, "params.numel")))

        dtype = get_dtype()
        self.m = torch.zeros(d, dtype=dtype)
        self.v = torch.zeros(d, dtype=dtype)

    def quantity(self, info: Info) -> torch.Tensor:
        """Gets the gradients.

        Args:
            info (Info): The job information object.

        Returns:
            torch.Tensor: The flat gradients.
        """
        grads_list: list[torch.Tensor] = []
        for p in info.model.params_.as_list:
            if p.grad is None:
                grads_list.append(torch.zeros_like(p))
            else:
                grads_list.append(p.grad)

        return torch.cat([g.reshape(-1) for g in grads_list]).detach()


class ParamStop(_BaseEMAStop):
    r"""Job to test the convergence based on the parameters' evolution.

    All stopping jobs have to agree and return True for the process to stop. Other
    jobs return None.

    Mathematically, estimates of the first and the second moments of the parameters'
    difference are stochastically computed, using exponential moving averages with the
    formula:

    .. math::
        m_i^{(t)} \gets \beta_i m_i^{(t-1)} + (1 - \beta_i) (\theta^{(t)} -
        \theta^{(t-1)})^i, \quad \hat{m}_i^{(t)} = \frac{m_i^{(t)}}{1 - \beta_i^t}.

    The `run` method returns true when for `min_consecutive` iterations:

    .. math::
        \vert \hat{m}_1^{(t)} \vert \leq \text{atol} + \text{rtol} \odot
        \sqrt{\hat{m}_2^{(t)}}.

    Please note the tolerances must both be non-negative and can represent element-wise
    tolerances or global tolerance common to all parameters.

    Attributes:
        atol (int | float): Absolute tolerance.
        min_consecutive (int): Minimum consecutive iterations to declare convergence.
        betas (tuple[int | float, int | float]): The forget parameters
        m (torch.Tensor): The first moment estimate of the log likelihood gain.
        v (torch.Tensor): The second moment estimate of the log likelihood gain.
        prev_loglik (torch.Tensor): The previous log likelihood value.
        n_consecutive (int): The number of iterations with convergence criterion met.
        stopped (bool): Indicator whether or not convergence has happened.
    """

    prev_params: torch.Tensor

    @wraps(_BaseEMAStop.__init__)
    def __init__(self, *args: Any, info: Info, **kwargs: Any):
        super().__init__(*args, **kwargs)

        d = info.model.params_.numel
        if isinstance(self.atol, torch.Tensor):
            check_consistent_size(((self.atol, 0, "atol"), (d, None, "params.numel")))
        if isinstance(self.rtol, torch.Tensor):
            check_consistent_size(((self.rtol, 0, "rtol"), (d, None, "params.numel")))

        dtype = get_dtype()
        self.m = torch.zeros(d, dtype=dtype)
        self.v = torch.zeros(d, dtype=dtype)

        self.prev_params = info.model.params_.as_flat_tensor

    def quantity(self, info: Info) -> torch.Tensor:
        """Gets the parameters' difference.

        Args:
            info (Info): The job information object.

        Returns:
            torch.Tensor: The flat parameters' difference.
        """
        new_params = info.model.params_.as_flat_tensor
        diff = new_params - self.prev_params
        self.prev_params = new_params
        return diff


class ValueStop(_BaseEMAStop):
    r"""Job to test the convergence based on log likelihood value.

    All stopping jobs have to agree and return True for the process to stop. Other
    jobs return None.

    Mathematically, estimates of the first and the second moments of the log likelihood
    difference are stochastically computed, using exponential moving averages with the
    formula:

    .. math::
        m_i^{(t)} \gets \beta_i m_i^{(t-1)} + (1 - \beta_i) \Bigl(
        \frac{\log \mathcal{L}(\theta^{(t)}) - \log \mathcal{L}(\theta^{(t-1)})}{n}
        \Bigr)^i, \quad \hat{m}_i^{(t)} = \frac{m_i^{(t)}}{1 - \beta_i^t}.

    The `run` method returns true when for `min_consecutive` iterations:

    .. math::
        \vert \hat{m}_1^{(t)} \vert \leq \text{atol} + \text{rtol}
        \sqrt{\hat{m}_2^{(t)}}.

    Please note the tolerances must be non-negative.

    Attributes:
        atol (int | float): Absolute tolerance.
        min_consecutive (int): Minimum consecutive iterations to declare convergence.
        betas (tuple[int | float, int | float]): The forget parameters
        m (torch.Tensor): The first moment estimate of the log likelihood gain.
        v (torch.Tensor): The second moment estimate of the log likelihood gain.
        prev_loglik (torch.Tensor): The previous log likelihood value.
        n_consecutive (int): The number of iterations with convergence criterion met.
        stopped (bool): Indicator whether or not convergence has happened.
    """

    prev_loglik: torch.Tensor

    @wraps(_BaseEMAStop.__init__)
    def __init__(self, *args: Any, info: Info, **kwargs: Any):
        super().__init__(*args, **kwargs)

        if isinstance(self.atol, torch.Tensor):
            check_consistent_size(((self.atol, 0, "atol"), (1, None, "1")))
        if isinstance(self.rtol, torch.Tensor):
            check_consistent_size(((self.rtol, 0, "rtol"), (1, None, "1")))

        dtype = get_dtype()
        self.m = torch.zeros(1, dtype=dtype)
        self.v = torch.zeros(1, dtype=dtype)

        self.prev_loglik = info.sampler.aux.logliks.mean()

    def quantity(self, info: Info) -> torch.Tensor:
        """Gets the log likelihood difference.

        Args:
            info (Info): The job information object.

        Returns:
            torch.Tensor: The log likelihood difference.
        """
        new_loglik = info.sampler.aux.logliks.mean()
        diff = new_loglik - self.prev_loglik
        self.prev_loglik = new_loglik
        return diff



import copy
from typing import Any

import torch
from pydantic import ConfigDict, validate_call

from ..typedefs._data import SampleData
from ..typedefs._defs import (
    Info,
    IntStrictlyPositive,
    Job,
    Metrics,
    Tensor2D,
    TensorCol,
    Trajectory,
)
from ..typedefs._params import ModelParams
from ..utils._checks import check_consistent_size, check_inf, check_nan
from ..utils._dtype import get_dtype


class PredictY(Job):
    r"""Job to predict longitudinal values.

    For every drawing of a random effect :math:`b`, this computes at the prediction
    times :math:`u` the values of the regression function given input data:

    .. math::
        h(u, b).

    The variable `u` is expected to be a matrix with the same number of rows as
    individuals, and the same number of columns as prediction times.

    Attributes:
        u (torch.Tensor): The matrix containing prediction times.
        pred_y (list[torch.Tensor]): The list of predicted longitudinal values.
    """

    u: torch.Tensor
    pred_y: list[torch.Tensor]

    @validate_call(config=ConfigDict(arbitrary_types_allowed=True))
    def __new__(cls, u: Tensor2D):
        """Creates the predicting job.

        Args:
            u (Tensor2D): The matrix containing prediction times.
        """
        return super().__new__(cls, u)

    def __init__(self, u: Tensor2D, info: Info):  # type: ignore
        """Initializes the predicting job.

        Args:
            u (Tensor2D): The matrix containing prediction times.
            info (Info): The job information object.

        Raises:
            ValueError: If u contains inf values.
            ValueError: If u contains NaN values.
            ValueError: If u has incompatible shape.
        """
        self.u = u.to(get_dtype())
        self.pred_y = []

        check_inf(((self.u, "u"),))
        check_nan(((self.u, "u"),))
        check_consistent_size(((self.u, 0, "u"), (info.data.size, None, "data.size")))

    def run(self, info: Info):
        """Appends the predicted longitudinal values.

        Args:
            info (Info): The job information object.
        """
        y = info.model.model_design.regression_fn(self.u, info.sampler.aux.psi)
        self.pred_y += [y[i] for i in range(y.size(0))]

    def end(self, metrics: Metrics, **_kwargs: Any):
        """Writes the predicted longitudinal values.

        Args:
            metrics (Metrics): The job metrics object.
        """
        metrics.pred_y = self.pred_y


class PredictSurvLogps(Job):
    r"""Job to predict survival log probability values.

    For every drawing of a random effect :math:`b`, this computes at the prediction
    times :math:`u` the values of the log survival probabilities given input data and
    conditionally to survival up to time :math:`c`:

    .. math::
        \log \mathbb{P}(T^* \geq u \mid T^* > c) = -\int_c^u \lambda(t) \, dt.

    When multiple transitions are allowed, :math:`\lambda(t)` is a sum over all
        possible transitions, that is to say if an individual is in the state :math:`k`
        from time :math:`t_0`, this gives:

        .. math::
            -\int_c^u \sum_{k'} \lambda^{k' \mid k}(t \mid t_0) \, dt.

    Please note this makes use of the Chasles property in order to avoid the computation
    of two integrals and make computations more precise.

    The variable `u` is expected to be a matrix with the same number of rows as
    individuals, and the same number of columns as prediction times.

    Attributes:
        u (torch.Tensor): The matrix containing prediction times.
        pred_surv_logps (torch.Tensor): The predicted survival log probabilities.
    """

    u: torch.Tensor
    pred_surv_logps: list[torch.Tensor]

    @validate_call(config=ConfigDict(arbitrary_types_allowed=True))
    def __new__(cls, u: Tensor2D):
        """Creates the predicting job.

        Args:
            u (Tensor2D): The matrix containing prediction times.
        """
        return super().__new__(cls, u)

    def __init__(self, u: Tensor2D, info: Info):  # type: ignore
        """Inits the predicting job.

        Args:
            u (Tensor2D): The matrix containing prediction times.
            info (Info): The job information object.

        Raises:
            ValueError: If u contains inf values.
            ValueError: If u contains NaN values.
            ValueError: If u has incompatible shape.
        """
        self.u = u.to(get_dtype())
        self.pred_surv_logps = []

        check_inf(((self.u, "u"),))
        check_nan(((self.u, "u"),))
        check_consistent_size(((self.u, 0, "u"), (info.data.size, None, "data.size")))

    def run(self, info: Info):
        """Computes and appends the survival log probabilities.

        Args:
            info (Info): The job information object.
        """
        sample_data = SampleData(
            info.data.x,
            info.data.trajectories,
            info.sampler.aux.psi,
            info.data.c,
            skip_validation=True,
        )
        surv_logps = info.model.compute_surv_logps(sample_data, self.u)

        self.pred_surv_logps += [surv_logps[i] for i in range(surv_logps.size(0))]

    def end(self, metrics: Metrics, **_kwargs: Any):
        """Writes the predicted log survival probabilities to metrics.

        Args:
            metrics (Metrics): The job metrics object.
        """
        metrics.pred_surv_logps = self.pred_surv_logps


class PredictTrajectories(Job):
    r"""Job to predict  trajectories.

    For every drawing of a random effect :math:`b`, this simulates the trajectories
    up to time `c_max` with a maximum length of `max_length` to avoid infinite loops.

    The variable `c_max` is expected to be a column vector with the same number of rows
    as individuals.

    Attributes:
        c_max (torch.Tensor): The maximum sampling (censoring) times.
        max_length (int): The max length of the trajectories.
        pred_trajectories (list[list[Trajectory]]): The predicted (sampled)
            trajectories.
    """

    c_max: torch.Tensor
    max_length: int
    pred_trajectories: list[list[Trajectory]]

    @validate_call(config=ConfigDict(arbitrary_types_allowed=True))
    def __new__(
        cls,
        c_max: TensorCol,
        max_length: IntStrictlyPositive = 10,
    ):
        """Creates the sampling predicting job.

        Args:
            c_max (TensorCol): The maximum sampling (censoring) times.
            max_length (IntStrictlyPositive, optional): The max length of the
                trajectories. Defaults to 10.
        """
        return super().__new__(cls, c_max, max_length)

    def __init__(  # type: ignore
        self,
        c_max: TensorCol,
        max_length: IntStrictlyPositive = 10,
        *,
        info: Info,
    ):
        """Inits the sampling predicting job.

        Args:
            c_max (TensorCol): The maximum sampling (censoring) times.
            max_length (IntStrictlyPositive, optional): The max length of the
                trajectories. Defaults to 10.
            info (Info): The job information object.

        Raises:
            ValueError: If c_max contains inf values.
            ValueError: If c_max contains NaN values.
            ValueError: If c_max has incompatible shape.
        """
        self.c_max = c_max.to(get_dtype())
        self.max_length = max_length
        self.pred_trajectories = []

        check_inf(((self.c_max, "c_max"),))
        check_nan(((self.c_max, "c_max"),))
        check_consistent_size(
            ((self.c_max, 0, "c_max"), (info.data.size, None, "data.size"))
        )

    def run(self, info: Info):
        """Predicts a trajectory given the current random effects drawing.

        Args:
            info (Info): The job information object.
        """
        for i in range(info.sampler.aux.psi.size(0)):
            sample_data = SampleData(
                info.data.x,
                info.data.trajectories,
                info.sampler.aux.psi[i],
                info.data.c,
                skip_validation=True,
            )

            self.pred_trajectories.append(
                info.model.sample_trajectories(
                    sample_data, self.c_max, max_length=self.max_length
                )
            )

    def end(self, metrics: Metrics, **_kwargs: Any):
        """Writes the predicted trajectories.

        Args:
            metrics (Metrics): The job metrics object.
        """
        metrics.pred_trajectories = self.pred_trajectories


class SwitchParams(Job):
    """Job to simulate different parameter values.

    This is useful when using the double Monte Carlo scheme of prediction.

    You can use this in conjunction with the `sample_params` method exposed in the
    `MultiStateJointModel` class.

    Examples:
        >>> params_list = my_model.sample_params(100)
        >>> my_model.do(pred_data, jobs=[PredictY(), SwitchParams(params_list)])
    """

    param_list: list[ModelParams]
    n_iterations_per_param: int
    n_params: int
    init_params: ModelParams

    @validate_call(config=ConfigDict(arbitrary_types_allowed=True))
    def __new__(
        cls, param_list: list[ModelParams], n_iterations_per_param: IntStrictlyPositive
    ):
        """Creates the job to switch between multiple parameters.

        Args:
            param_list (list[ModelParams]): The list of parameters.
            n_iterations_per_param (IntStrictlyPositive): The number of iterations to
                execute before switching.
        """
        return super().__new__(cls, param_list, n_iterations_per_param)

    def __init__(  # type: ignore
        self,
        param_list: list[ModelParams],
        n_iterations_per_param: IntStrictlyPositive,
        info: Info,
    ):
        """Initializes the job to switch between multiple parameters.

        Args:
            param_list (list[ModelParams]): The list of parameters.
            n_iterations_per_param (IntStrictlyPositive): The number of iterations to
                execute before switching.
            info (Info): The job information object.
        """
        self.param_list = param_list
        self.n_iterations_per_param = n_iterations_per_param
        self.n_params = len(param_list)

        self.init_params = copy.deepcopy(info.model.params_)

    def run(self, info: Info):
        """Executes the switches once every `n_iterations_per_param`.

        Args:
            info (Info): The job information object.
        """
        if info.iteration % self.n_iterations_per_param == 0:
            info.model.params_ = self.param_list[
                (info.iteration // self.n_iterations_per_param) % self.n_params
            ]

    def end(self, info: Info, **_kwargs: Any):
        """Restore default parameters.

        Args:
            info (Info): The job information object.
        """
        info.model.params_ = self.init_params
